/**==========================================================================

	PC-8031 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2024 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	last modified  Apr,27,2024.

;==========================================================================*/
#include "../PC-8001/PC-8001.h"
#ifdef __APPLE__
#include <string.h>
#endif
#ifdef __cplusplus
extern "C"
{
#endif
extern USHORT read_short(LPBYTE ptr);
extern DWORD read_quad(LPBYTE ptr);
#define export
/**--------------------------------------------------------------------------
 **		Disk Drive
 **------------------------------------------------------------------------*/
static void _drv_eject(LPFDUDRIVE drv) {
	if (!drv->media_inserted)
		return;
	if (drv->media == NULL) {
		drv->media_inserted = FALSE;
		return;
	}
	drv->media->free(drv->media);
	drv->media_inserted = FALSE;
	drv->media = NULL;
}
static BOOL _drv_insert(LPFDUDRIVE drv, LPFLOPPYDISK media) {
	if (drv->media_inserted)
		return FALSE;
	if (drv->media != NULL) {
		drv->media_inserted = TRUE;
		return FALSE;
	}
	drv->media_inserted = TRUE;
	drv->media = media;
	return TRUE;
}
static char _drv_get_state(LPFDUDRIVE drv) {
	return drv->missing_unit || !drv->media_inserted ? '-' : drv->state;
}
static void _drv_reset(LPFDUDRIVE drv) {
	drv->status = 0;
	drv->state	= '*';
	drv->busy	= DRVBSY_IDOL;
}
static void _drv_init_drive(LPFDUDRIVE drv, BYTE unit_no) {
	drv->reset			= _drv_reset;
	drv->eject			= _drv_eject;
	drv->insert			= _drv_insert;
	drv->get_state		= _drv_get_state;
	drv->unit_no		= unit_no;
	drv->media_inserted = FALSE;
	drv->media			= NULL;
	drv->reset(drv);
}
/**--------------------------------------------------------------------------
 **		Disk Drives
 **------------------------------------------------------------------------*/
static int _fdd_seek(LPFDUDRIVES fdd, int unit, int track) {
	LPFDUDRIVE drive = &fdd->drives[unit];
	int seek_interval = 1;
	if (drive->missing_unit /*|| !drive->media_inserted*/) {
		drive->status = (unit & FDC_ST0_US_MASK) | FDC_ST0_SE | FDC_ST0_NR;// | FDC_ST0_IC_ABNORM_TERM;
	} else {
		drive->status = (unit & FDC_ST0_US_MASK) | FDC_ST0_SE;
		drive->track = (BYTE)track;
		seek_interval = fdd->owner->fdc.step_rate_time;
	}
	drive->state = 'S';
	int diff = trkdiff(track, ((int)drive->id.C));
	return (seek_interval * diff);		// 1..16ms
}

static void _fdd_sense_interrupt_status(LPFDUDRIVES fdd) {
	LPFDU765A fdc = &fdd->owner->fdc;
	for (int i = 0; i < FDC_MAX_UNIT_CNT; ++i) {
		LPFDUDRIVE drive = &fdd->drives[i];
		if (drive->status != 0) {
			fdc->result.push0(&fdc->result, drive->status);		// ST0
			fdc->result.push(&fdc->result, drive->track);		// PCN
			drive->status = 0;
			drive->state = 'Q';
			return;
		}
	}
	fdc->result.push0(&fdc->result, FDC_ST0_IC_INVALID_CMD);	// ST0
	fdc->result.push(&fdc->result, 0);							// PCN
}

static void _fdd_sense_drive_status(LPFDUDRIVES fdd, int unit, int head) {
	LPFDU765A fdc = &fdd->owner->fdc;
	LPFDUDRIVE drive = &fdd->drives[unit];
	BYTE ST3 = (unit & FDC_ST3_US_MASK);
	if (drive->missing_unit)
		ST3 |= FDC_ST3_FT;
	else if (drive->media_inserted) {
		ST3 |= (FDC_ST3_RY | FDC_ST3_TS | (drive->track == 0 ? FDC_ST3_T0 : 0) | (drive->media->header->readonly == 0x10 ? FDC_ST3_WP : 0));
	} else
		ST3 |= FDC_ST3_TS;
	drive->state = 'Q';
	fdc->result.push0(&fdc->result, ST3);					// ST3
}

static int _fdd_read_data(LPFDUDRIVES fdd, int unit, LPFDUCOMMAND cmd, int multrk_cnt) {
	LPFDU765A fdc = &fdd->owner->fdc;
	LPFDUDRIVE drive = &fdd->drives[unit];
	if (multrk_cnt == 0) {
		//BYTE c = (cmd->id.C << 1) + cmd->id.H;
		drive->id = cmd->id;
	}
	drive->data_buff_index = 0;
	drive->data_size = 256;
	drive->remain_cnt = drive->data_size;
	memset(drive->data_buff, 0xff, drive->data_size);

	if (drive->missing_unit || !drive->media_inserted) {
		fdc->result.ST0 |= FDC_ST0_IC_ABNORM_TERM;
		fdc->result.ST1 |= FDC_ST1_MA | FDC_ST1_EN | FDC_ST1_DE | FDC_ST1_ND;
		return -1;
	}
	LPD88SECTOR_EX sector = drive->media->find_sector(drive->media, &drive->id);
	if (sector == NULL) {
		fdc->result.ST0 |= FDC_ST0_IC_ABNORM_TERM;
		fdc->result.ST1 |= FDC_ST1_MA | FDC_ST1_EN | FDC_ST1_DE | FDC_ST1_ND;
		return -1;
	} else {
#if _DEBUG
		n80device.logger("     {READ SECTOR} [C:%d, H:%d, R:%d, N:%d]\n",
			drive->id.C, drive->id.H, drive->id.R, drive->id.N);
#endif
		// move id to next
		if (drive->id.R < 16) {
			drive->id.R++;
			//} else if (drive->id.H == 0) {
			//	drive->id.H = 1;
			//	drive->id.R = 1;
		} else {
			drive->id.C++;
			drive->id.H = 0;
			drive->id.R = 1;
		}
		fdc->result.id = drive->id;
	}
	drive->state = 'R';
	memcpy(drive->data_buff, sector->data, drive->data_size);
	return drive->data_size;
}
static void _fdd_reset(LPFDUDRIVES fdd) {
	for (int u = 0; u < FDC_MAX_UNIT_CNT; ++u) {
		LPFDUDRIVE drv = &fdd->drives[u];
		drv->reset(drv);
	}
}
export void _fdu_init_drives(LPFDUDRIVES fdd, LPFDU owner) {
	fdd->owner	= owner;
	fdd->seek	= _fdd_seek;
	fdd->snsint = _fdd_sense_interrupt_status;
	fdd->snsdrv = _fdd_sense_drive_status;
	fdd->read	= _fdd_read_data;
	fdd->reset	= _fdd_reset;
	fdd->drives_cnt = 2;
	for (int u = 0; u < FDC_MAX_UNIT_CNT; ++u) {
		LPFDUDRIVE drv = &fdd->drives[u];
		drv->owner = fdd;
		drv->init = _drv_init_drive;
		drv->missing_unit = u >= fdd->drives_cnt;
		drv->init(drv, u);
	}
}
#ifdef __cplusplus
}	// extern "C"
#endif
