//
//  ViewController.swift
//  N80X
//
//  Created by hiroshi on 2021/08/20.
//
import Cocoa
import Metal
import MetalKit
import GameController

struct Vertex {
    var position : vector_float2
    var pixel : vector_float2
}

class ViewController: NSViewController {
    private let device = MTLCreateSystemDefaultDevice()!
    private var mtkView : MTKView! = nil
    private var commandQueue: MTLCommandQueue! = nil
    private var renderPassDescriptor: MTLRenderPassDescriptor! = nil
    private var bufferPosition: MTLBuffer! = nil
    private var bufferColor: MTLBuffer! = nil
    private var renderPipelineState: MTLRenderPipelineState! = nil
    private var textureDescriptor : MTLTextureDescriptor! = nil
    private let vertexData = [
        Vertex(position: [-1, -1], pixel: [0, 0]),
        Vertex(position: [1, -1], pixel: [1, 0]),
        Vertex(position: [-1, 1], pixel: [0, 1]),
        Vertex(position: [1, 1], pixel: [1, 1])
    ]
    fileprivate var n80em:n80em! = nil
    fileprivate var n80view:N80View! = nil

    override open func loadView() {
        super.loadView()
        self.n80view = self.view as? N80View
    }

    override func viewDidLoad() {
        super.viewDidLoad()

		let version = Bundle.main.object(forInfoDictionaryKey: "CFBundleShortVersionString") as? String
		let build = Bundle.main.object(forInfoDictionaryKey: "CFBundleVersion") as? String

        // Do any additional setup after loading the view.
        if let window = NSApplication.shared.windows.first {
            window.delegate = self
            window.title = "N80X " + (version ?? "") + (build ?? "")
        }
        // initialize n80em object and get it
        let appDelegate:AppDelegate = NSApplication.shared.delegate as! AppDelegate
        appDelegate.req_init_n80()
        self.n80em = appDelegate.N80
        
        self.regist_halt_callback()
        self.regist_gckey_callback()

        if self.n80em!.main_rom_ready {
            if !n80device.fdu.rom_ready {
                appDelegate.menuOpenDiskRom.isHidden = false;
            } else {
                appDelegate.menuDrive1.isHidden = false
                appDelegate.menuDrive2.isHidden = false
            }
            self.n80em!.start() // n80em prepare start
            // z80 thread start
            var _ = Thread.detachNewThreadSelector(#selector(ViewController.clock), toTarget: self, with: nil)
        } else {
            appDelegate.menuEmu.isEnabled = false
            appDelegate.callback = { () -> Bool in
                return self.load_main_rom_callback()
            }
        }
        self.prepareMetal()     // prepare metal objects
    }
    
    private func regist_gckey_callback() {
        // register event on Keyboard connection detected, then capture keys!
        NotificationCenter.default.addObserver(forName: NSNotification.Name.GCKeyboardDidConnect, object: nil, queue: OperationQueue.main) { (notification) in
            if let keyboard = GCKeyboard.coalesced?.keyboardInput {
                keyboard.keyChangedHandler = {
                    (keyboard, key, keyCode, pressed) in
                    if pressed {
                        self.n80em.n80kb.kb_keydown(keyCode)
                    } else {
                        self.n80em.n80kb.kb_keyup(keyCode)
                    }
                }
            }
        }
    }
    
    private func regist_halt_callback() {
        // register illegal instruction detect error handler
        self.n80em.halt_callback = {
            (halt) in
            var msg:String = "";
            switch halt {
            case 1:
                msg = "detect HALT instruction"
            case 2:
                msg = "detect illegal instruction"
            default:
                msg = "detect Unimplemented instruction"
            }
            let alert = NSAlert()
            alert.alertStyle = .warning
            alert.informativeText = msg
            alert.messageText = "Z80 CODE ERROR"
            alert.addButton(withTitle: "OK")
            let _ = alert.runModal()
        }
    }
    
    private func load_main_rom_callback() -> Bool {
        if !self.n80em!.main_rom_ready {
            if !self.n80em!.req_load_main_rom() {
                let alert = NSAlert()
                alert.alertStyle = .warning
                alert.informativeText = "can not load Main ROM file"
                alert.messageText = "N80 Error"
                alert.addButton(withTitle: "OK")
                let _ = alert.runModal()
                return false
            }
            self.n80em!.main_rom_ready = true;
            self.n80em!.start()     // n80em prepare start
            // z80 thread start
            var _ = Thread.detachNewThreadSelector(#selector(ViewController.clock), toTarget: self, with: nil)
            return true
        }
        return false
    }

    @objc open func clock() {
        self.n80em.enter_z80_thread()
    }
    //-------------------------------------------------------------------------
    //    initialize Metal
    //-------------------------------------------------------------------------
    func prepareMetal() {
        let (w,h) = (view.frame.width,view.frame.height)
        let mtkViewFrame = CGRect(x: w/2 - 2*h/3, y: 0,width: 4*h/3, height: h)
        self.mtkView = MTKView(frame: mtkViewFrame, device: device)
        self.mtkView.device = device
        self.mtkView.framebufferOnly = true
        self.mtkView.preferredFramesPerSecond = 60
        self.mtkView.colorPixelFormat = .bgra8Unorm
        self.mtkView.clearColor = MTLClearColor(red: 0, green: 0, blue: 0, alpha: 1)
        self.mtkView.delegate = self
        n80view!.subviews.insert(mtkView, at: 0)

        // init MTLCommandQueue
        commandQueue = device.makeCommandQueue()
        
        // create Render Pass Descriptor
        // set behavor of this Render Pass when it will be executed
        // set background color to black
        renderPassDescriptor = MTLRenderPassDescriptor()
        renderPassDescriptor.colorAttachments[0].loadAction = MTLLoadAction.clear
        renderPassDescriptor.colorAttachments[0].storeAction = MTLStoreAction.store
        renderPassDescriptor.colorAttachments[0].clearColor = MTLClearColorMake(0.0, 0.0, 0.0, 1.0)
        // create vertex buffer
        let size = vertexData.count * MemoryLayout<Vertex>.size
        bufferPosition = device.makeBuffer(bytes: vertexData, length: size)
        // regist shader
        guard let library = device.makeDefaultLibrary() else {fatalError()}
        let descriptor = MTLRenderPipelineDescriptor()
        descriptor.vertexFunction = library.makeFunction(name: "myVertexShader")
        descriptor.fragmentFunction = library.makeFunction(name: "video_fragment")
        descriptor.colorAttachments[0].pixelFormat = .bgra8Unorm
        // create Render pipe-line state
        renderPipelineState = try! device.makeRenderPipelineState(descriptor: descriptor)
        // create texture descriptor (RGB[bitmap] -> BRG[metal])
        let appDelegate:AppDelegate = NSApplication.shared.delegate as! AppDelegate
        let bmpimg = appDelegate.BmpImg
        self.textureDescriptor = MTLTextureDescriptor.texture2DDescriptor(pixelFormat: MTLPixelFormat.rgba8Unorm, width: Int(bmpimg.width), height: Int(bmpimg.height), mipmapped: false)
    }
    func makeTexture() -> MTLTexture {
        let appDelegate:AppDelegate = NSApplication.shared.delegate as! AppDelegate
        let bmpimg = appDelegate.BmpImg
        let tex = device.makeTexture(descriptor: self.textureDescriptor)
        let rgn = MTLRegionMake2D(0, 0, Int(bmpimg.width), Int(bmpimg.height))
        // set off-screen bitmap to texture
        tex?.replace(region: rgn, mipmapLevel: 0, slice: 0, withBytes: bmpimg.bitmap, bytesPerRow: Int(bmpimg.bytesPerRow), bytesPerImage: Int(bmpimg.bytesPerRow*bmpimg.height))
        return tex!
    }
}
extension ViewController : NSWindowDelegate {
    func windowDidResize(_ notification: Notification) {
        if self.mtkView != nil {
            let (w,h) = (view.frame.width,view.frame.height)
            let mtkViewFrame = CGRect(x: w/2 - 2*h/3, y: 0,width: 4*h/3, height: h)
            self.mtkView.frame = mtkViewFrame
        }
    }
}
extension ViewController : MTKViewDelegate {
    
    public func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
    }
    
    public func draw(in view: MTKView) {
        // crate texture
        let tex = makeTexture()
        // purchase drawable
        guard let drawable = view.currentDrawable else {fatalError()}
        renderPassDescriptor.colorAttachments[0].texture = drawable.texture
        // create Command Buffer
        guard let cmdBuffer = commandQueue.makeCommandBuffer() else {fatalError()}
        // create Encoder
        let encoder = cmdBuffer.makeRenderCommandEncoder(descriptor: renderPassDescriptor)!
        encoder.setRenderPipelineState(renderPipelineState)
        // send vertex points array to encoder
        encoder.setVertexBuffer(bufferPosition, offset: 0, index: 0)
        encoder.setFragmentTexture(tex, index: 0)
        // draw rectangle
        encoder.drawPrimitives(type: MTLPrimitiveType.triangleStrip, vertexStart: 0, vertexCount: 4)
        // end encoding
        encoder.endEncoding()
        // entry drawable
        cmdBuffer.present(drawable)
        // commit command buffer
        cmdBuffer.commit()
        //cmdBuffer.waitUntilCompleted()
        
        // request next off-screen bitmap painting to emulator
        //self.n80em.refleshScreen()
    }
}

