//
//  n80em.m
//
//  Copyright (C) 1993-2019 Hiroshi Kanemoto.
//    Released under the MIT license
//    http://opensource.org/licenses/mit-license.php
//
//     Created:    2012/08/22    H.Kanemoto
//    Modified:    2016/04/05    H.Kanemoto    ported from MaxOS X
//    Modified:    2019/05/12    H.Kanemoto    support METAL
//
#import "N80X-Swift.h"
#import "n80em.h"
#import "n80sound.h"
#import "n80graph.h"
#import "n80keyboard.h"
#import "n80config.h"
#import <stdatomic.h>
#include <mach/mach_time.h>
#include <AudioUnit/AudioUnit.h>
#include <stdio.h>
//-----------------------------------------------------------------------------
//    Emulator C Functions
//-----------------------------------------------------------------------------
#define NOTDONE     0
#define DONE        1
#define STDSIZE     (32576)
#define MAXSIZE     (32768)
#define SP_VAL      (0x7f3e)    // N80 file: offset to stack pointer
#define PC_VAL      (0xff3d)    // N80 file: offset to return address
#define IMGTYP_BAD  (-1)
#define IMGTYP_N80  (0)
#define IMGTYP_CMT  (1)
#define IMGTYP_T88  (2)
#define IMGTYP_D88  (3)

export N80LX        n80lx;
static n80em        *emu;
//-----------------------------------------------------------------------------
//    export "C" functions of bridge Objective-C for n80lx emulator module
//-----------------------------------------------------------------------------
static void _n80_beep_on(void) {
    [emu.n80snd beep_on];
}
static void _n80_beep_off(void) {
    [emu.n80snd beep_off];
}
static void _n80_sound_on(USHORT freq, int ch) {
    [emu.n80snd sound_on:freq ch:(int)ch];
}
static void _n80_sound_off(void) {
    [emu.n80snd sound_off];
}
static void _n80_chg_bltfunc(int mode) {
    // mode: 0 = 40x20, 1 = 40x25, 2 = 80x20, 3 = 80x25
    [emu.n80gr chg_bltfunc:mode];
}
static void _n80_chg_palette(bool interrace) {
    [emu.n80gr chg_palette:interrace];
}
static void _n80_vram_enter(void) {
    [emu.n80gr lock];
}
static void _n80_vram_leave(void) {
    n80lx.vram_sync = 0;
    [emu.n80gr unlock];
}
export void _n80_init_snd(void) {
    n80device.snd.beep_on      = _n80_beep_on;
    n80device.snd.beep_off     = _n80_beep_off;
    n80device.snd.sound_on     = _n80_sound_on;
    n80device.snd.sound_off    = _n80_sound_off;
}
export void _n80_init_blt(void) {
    n80device.blt.chg_crt_mode = _n80_chg_bltfunc;
    n80device.blt.chg_palette  = _n80_chg_palette;
}
//-----------------------------------------------------------------------------
//    supplement function
//-----------------------------------------------------------------------------
static double system_tick_time(void) {
    struct timeval time;
    gettimeofday(&time, NULL);
    return time.tv_sec + (double) time.tv_usec*1e-6;
}
//-----------------------------------------------------------------------------
//    Emulator Objective-C Class
//-----------------------------------------------------------------------------
@implementation n80em

-(id)init:(DblBuffBmpImage*)bmp data:(n80config_data*)data {
    if (self = [super init]) {
        emu = self;
        // init member variables
        self->dblbmp = bmp;
        self.running  = false;
        self.terminating = false;    // leave pause before terminating
        self.config_mode = false;
        self.halt_callback = nil;

        // N80 structure clean
        memset(&n80lx, 0, sizeof(n80lx));
        
        // PC-8001 Device Initialize
        _n80_device_initialize(&n80device, _n80_vram_enter, _n80_vram_leave);
        
        // load font file
        [self loadFontFile];
        
        // load rom file
        self.main_rom_ready = [self loadROMFile:data];

        pauseZ80sync = [[NSConditionLock alloc] initWithCondition:NOTDONE];

        // create/initialize sound module
        self->sound = [[n80sound alloc] initWithN80:self];
        // create/initialize graph module
        self->graph = [[n80graph alloc] initWithN80:self];
        // create/initialize keyboard module
        self->keyboard = [[n80keyboard alloc] initWithN80:self];
        // create/initialize config module
        self->config = [[n80config alloc] initWithN80:self config:data];
        
        _n80_chg_palette(data.interlace);
        [self->config reflect_menu_state];
    }
    return self;
}

-(void)dealloc {
    NSLog(@"n80em dealloc");
    [self req_dispose];
}
//-----------------------------------------------------------------------------
//    property methods
//-----------------------------------------------------------------------------
-(N80LX const *)getN80 {
    return &n80lx;
}

-(n80sound const *)getN80Sound {
    return self->sound;
}

-(n80graph const *)getN80Graph {
    return self->graph;
}

-(n80keyboard const *)getN80Keyboard {
    return self->keyboard;
}

-(n80config const *)getN80Config {
    return self->config;
}

-(DblBuffBmpImage const *)getBmpimg {
    return self->dblbmp;
}
//-----------------------------------------------------------------------------
//    N80 sync
//-----------------------------------------------------------------------------
-(void)suspend {
    [self.n80snd suspend];
    [self.n80gr suspend];
    self.running  = false;
    self->state = 0;
}

-(void)resume {
    if ([self.n80snd resume]) {
        [self start];
    }
}

- (void)start {
    self.running  = true;
    self.pause = false;
    self->state = 0;
    [self.n80gr start];
}
//-----------------------------------------------------------------------------
//    N80 request
//-----------------------------------------------------------------------------
- (void)req_dispose {
    atomic_store(&self->request_cpu_signal, 1);
    if (self.running) {
        self.running = false;
    }
    // stop sound unit
    [self.n80snd reqDispose];
    if (n80device.cmt._buff != NULL) {
        free(n80device.cmt._buff);
        n80device.cmt._buff = NULL;
    }
    // release graph unit
    [self.n80gr reqDispose];
    self->dblbmp = nil;
    // release floppy media
    n80device.fdu.fdd.drives[0].eject(&n80device.fdu.fdd.drives[0]);
    n80device.fdu.fdd.drives[1].eject(&n80device.fdu.fdd.drives[1]);
}
- (void)req_reset {
    if (self.config_mode)
        return;
    int expected = 0;
    while (!atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 2)) {
        expected = 0;
    }
}
- (void)req_toggle_pcg {
    n80device.pcg_mode ^= 0x01;
    n80device.attr.invalid();
    [self.n80cfg reflect_menu_state];
}
-(void)req_enter_config {
    AppDelegate *app = [[NSApplication sharedApplication] delegate];
    [self enterPause];
    self.config_mode = true;
    [app.menuEmu setEnabled:false];
    [app.menuFile setEnabled:false];
    [self.n80cfg enter_config];
}
-(void)req_leave_config {
    AppDelegate *app = [[NSApplication sharedApplication] delegate];
    [self.n80cfg leave_config];
    [self leavePause:false];
    self.config_mode = false;
    [app.menuEmu setEnabled:true];
    [app.menuFile setEnabled:true];
}
-(void)req_eject:(int)unit {
    n80device.fdu.fdd.drives[unit].eject(&n80device.fdu.fdd.drives[unit]);
}

#if TARGET_OS_OSX
- (void)req_open_image_file {
    if (self.config_mode)
        return;

    [self enterPause];
    FILE *fp = NULL;
    int image_type = IMGTYP_BAD;
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *allowedFileTypes = [NSArray arrayWithObjects:@"N80",@"'N80 Memory Image File'",@"CMT",@"'Casset Image File'",@"T88",@"'Casset Image File'",nil];
    [openPanel setAllowedFileTypes:allowedFileTypes];
    NSInteger pressedButton = [openPanel runModal];
    if (pressedButton == NSModalResponseOK) {
        NSString *path = [[openPanel URL] path];
        if ([path hasSuffix:@"N80"] || [path hasSuffix:@"n80"])
            image_type = IMGTYP_N80;
        else if ([path hasSuffix:@"T88"] || [path hasSuffix:@"t88"])
            image_type = IMGTYP_T88;
        else
            image_type = IMGTYP_CMT;
        // open N80/CMT/T88 image file
        const char *fn = [path fileSystemRepresentation];
        fp = fopen(fn,"r");
    }
    if (fp != NULL) {
        // get image file length
        fseek(fp, 0, SEEK_END);
        long fsiz = ftell(fp);
        rewind(fp);

        if (fsiz == 0)
            ;
        else if (image_type == IMGTYP_N80) {    // N80 File
            USHORT spval;
            
            if (fsiz >= STDSIZE) {
                fsiz = (fsiz > MAXSIZE) ? MAXSIZE : fsiz;
                fread(n80device.mem._ram, 1, fsiz, fp);
                fseek(fp, SP_VAL, SEEK_SET);
                fread(&spval, 2, 1, fp);
                n80device.cpu.r.x.PC = PC_VAL;
                n80device.cpu.r.x.SP = spval;
            }
        } else {
            if (n80device.cmt._buff != NULL) {
                free(n80device.cmt._buff);
                n80device.cmt._buff = NULL;
                n80device.cmt._size = 0;
            }
            n80device.cmt._buff = malloc(fsiz);
			if (n80device.cmt._buff == NULL) {
				;
            } else if (image_type == IMGTYP_CMT) {
                size_t len = fread(n80device.cmt._buff, 1, fsiz, fp);
                n80device.cmt._size  = (int) len;
                n80device.cmt._index = 0;

            } else if (image_type == IMGTYP_T88) {
				const char* header = "PC-8801 Tape Image(T88)";
				char buff[23 + 1];
                n80device.cmt._size  = (int)fsiz;
				for (int index = 0; index < fsiz;) {
					// file header check
					if (fread(buff, 1, 24, fp) != 24)
						break;
					if (strcmp(header, buff) != 0)
						break;
					// skip version tag
					if (fread(buff, 1, 6, fp) != 6)
						break;
					while (true) {
						// tag + siz read
						if (fread(buff, 1, 4, fp) != 4)
							break;
						int tag_siz = 0;
						tag_siz += ((BYTE)buff[3]);
						tag_siz <<= 8;
						tag_siz += ((BYTE)buff[2]);
						// blank/space/mark tags
						if ((buff[0] == 0x00 && buff[1] == 0x01) ||
							(buff[0] == 0x02 && buff[1] == 0x01) ||
							(buff[0] == 0x03 && buff[1] == 0x01)) {
							if (fread(buff, 1, tag_siz, fp) != tag_siz)
								break;
						// data tag
						} else if (buff[0] == 0x01 && buff[1] == 0x01) {
							// area time/area length/data length/data type
							if (fread(buff, 1, 12, fp) != 12)
								break;
							int data_siz = 0;
							data_siz += ((BYTE)buff[9]);
							data_siz <<= 8;
							data_siz += ((BYTE)buff[8]);
							// check buffer size
							if ((index + data_siz) > (int)n80device.cmt._size)
								break;
							// read data
							if (fread(&n80device.cmt._buff[index], 1, data_siz, fp) != data_siz)
								break;
							index += data_siz;
						// end tag
						} else if (buff[0] == 0x00 && buff[1] == 0x00) {
							n80device.cmt._size = index;
							n80device.cmt._index = 0;
							break;
						// invalid tag
						} else
							break;
					}
					break;
				}
            }
        }
        fclose(fp);
    }
    [self leavePause:false];
}
#else
- (void)reqOpenFile:(bytecoder*)data isN80:(BOOL)isN80File {
	[pauseZ80sync lock];
	int expected = 0;
	while (!atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 3)) {
		expected = 0;
	}
	//OSAtomicCompareAndSwap32(0, 3, &n80lx.request_disp_signal);
	
	if (isN80File) {	// N80 File
		USHORT spval;
		
		int fsiz = data.Length;
		if (fsiz >= STDSIZE) {
			fsiz = (fsiz > MAXSIZE) ? MAXSIZE : fsiz;
			[data moveptrto:0];
			[data pull:&n80device.mem.map._ram[0] Length:fsiz];
			[data moveptrto:SP_VAL];
			[data pull_le_short:&spval];
			n80lx.cpu.r.x.PC = PC_VAL;
			n80lx.cpu.r.x.SP = spval;
		}
	} else {			// CMT File
		if (n80device.cmt._buff != NULL) {
			free(n80device.cmt._buff);
			n80device.cmt._buff = NULL;
		}
		n80device.cmt._buff = malloc(data.Length);
		[data moveptrto:0];
		[data pull:n80device.cmt._buff Length:data.Length];
		n80device.cmt._size  = (int) data.Length;
		n80device.cmt._index = 0;
	}
	[pauseZ80sync unlock];
}
#endif
//-----------------------------------------------------------------------------
//    Initialize Emulator
//-----------------------------------------------------------------------------
-(BOOL)loadFontFile {
    long len = [self loadFileToBuffer:"PC-8001.FON" Size:1024*4 Buffer:n80device.font];
    if (len == 1024*4) {
        // copy font rom data to PCG ram
        for (int i = 0; i < 2048; ++i)
            n80device.pcg[i] = n80device.font[i];
        return true;
    }
    return false;
}

-(BOOL)loadROMFile:(n80config_data*)data {
    //return [self loadResFile:@"PC-8001.ROM" Size:1024*24 Buffer:&n80device.mem._rom[0]];
    if (![self loadMainROMFile:data]) {
        return false;
    }
    [self loadDiskROMFile:data];
    return true;
}

-(BOOL)loadMainROMFile:(n80config_data*)data {
    long len = 0;
    while (1) {
        char *fn = &data->slot[0][0];
        if (strlen(fn) > 0) {
            len = [self loadFileToBuffer:fn Size:1024*32 Buffer:n80device.mem._arr];
            if (len > 0)
                break;
        }
        fn = "PC-8001.ROM";
        len = [self loadFileToBuffer:fn Size:1024*32 Buffer:n80device.mem._arr];
        if (len > 0)
            break;
        return false;
    }
    if (len > 1024*24)
        n80device.param.attach_4th_rom = true;
    return true;
}

-(BOOL)req_load_main_rom {
    BOOL result = false;
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *allowedFileTypes = [NSArray arrayWithObjects:@"ROM",@"'Main ROM File'",nil];
    [openPanel setAllowedFileTypes:allowedFileTypes];
    NSInteger pressedButton = [openPanel runModal];
    if (pressedButton == NSModalResponseOK) {
        NSString *path = [[openPanel URL] path];
        long len = [self loadPathFile:path Size:32*1024 Buffer:n80device.mem._rom];
        if (len > 0) {
            if (len > 24*1024)
                n80device.param.attach_4th_rom = true;
            result = true;
        }
    }
    return result;
}

-(BOOL)loadDiskROMFile:(n80config_data*)data {
    long len = 0;
    char *fn = &data->slot[2][0];
    if (strlen(fn) > 0) {
        len = [self loadFileToBuffer:fn Size:1024*8 Buffer:n80device.fdu.mem._rom];
    }
    n80device.fdu.set_rom_ready(&n80device.fdu, len > 0);
    return len > 0;
}

-(BOOL)req_load_disk_rom {
    BOOL result = false;

    [self enterPause];
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *allowedFileTypes = [NSArray arrayWithObjects:@"ROM",@"'Disk ROM File'",nil];
    [openPanel setAllowedFileTypes:allowedFileTypes];
    NSInteger pressedButton = [openPanel runModal];
    if (pressedButton == NSModalResponseOK) {
        NSString *path = [[openPanel URL] path];
        long len = [self loadPathFile:path Size:8*1024 Buffer:n80device.fdu.mem._rom];
        if (len > 0) {
            result = true;
            n80device.fdu.set_rom_ready(&n80device.fdu, true);
        }
    }
    [self leavePause:false];
    return result;
}

-(long)loadFileToBuffer:(const char*)file_name Size:(int)size Buffer:(LPBYTE)buffer {
    size_t l = strlen(file_name);
    NSData *byte = [[NSData alloc] initWithBytesNoCopy:(void *)file_name length:l freeWhenDone:false];
    NSString *nsfn = [[NSString alloc] initWithData:byte encoding:NSShiftJISStringEncoding];
    // find rom file from App local path
    NSString *defPath = [[NSBundle mainBundle] pathForResource:nsfn ofType:nil];
    long len = [self loadPathFile:defPath Size:size Buffer:buffer];
// ~/配下のファイルを絶対パス(/Users/xxxxx/)で指定してもファイルアクセスの権限がないと言って怒られる、
// chmon 777 にしてもダメ。ホームディレクトリなのに、、、理由がわかる方ご教示頂けると助かります。
//    if (len <= 0) {
//        // find rom file from User default path
//        char path[512];
//        strcpy(path, "~/");
//        strcat(path, file_name);
//        NSData *byte2 = [[NSData alloc] initWithBytesNoCopy:(void *)path length:strlen(path) freeWhenDone:false];
//        NSString *nsfn2 = [[NSString alloc] initWithData:byte2 encoding:NSShiftJISStringEncoding];
//        len = [self loadPathFile:nsfn2 Size:size Buffer:buffer];
//        if (len <= 0)
//            return result;
//    }
    return len;
}

-(long)loadPathFile:(NSString*)path Size:(int)size Buffer:(LPBYTE)buffer {
    long result = -1;

    FILE *fp = NULL;
    const char *fn = [path fileSystemRepresentation];
    fp = fopen(fn, "rb");
    if (fp == NULL)
        return result;

    // get file length
    fseek(fp, 0, SEEK_END);
    long siz = ftell(fp);
    rewind(fp);
    
    if (siz > size)
        ;
    else {
        // read data from resource file
        size_t len = fread(buffer, 1, size, fp);
        if (siz == len)
            result = (long)len;
    }
    fclose(fp);
    return result;
}
static void d88_image_release(LPFLOPPYDISK media) {
    if (media != NULL) {
        if (media->image_raw_data != NULL)
            free(media->image_raw_data);
        free(media);
        NSLog(@"release D88 media");
    }
}
-(BOOL)req_open_d88_file:(int)unit {
    BOOL result = false;

    [self enterPause];
    FILE *fp = NULL;
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *allowedFileTypes = [NSArray arrayWithObjects:@"D88",@"'D88 Disk Image File'",nil];
    [openPanel setAllowedFileTypes:allowedFileTypes];
    NSInteger pressedButton = [openPanel runModal];
    if (pressedButton == NSModalResponseOK) {
        // open D88 image file
        NSString *path = [[openPanel URL] path];
        const char *fn = [path fileSystemRepresentation];
        fp = fopen(fn, "rb");
        if (fp != NULL) {
            // get D88 file length
            fseek(fp, 0, SEEK_END);
            long siz = ftell(fp);
            rewind(fp);
            
            if (siz > 0) {
                // create and initialize floppy media
                LPFLOPPYDISK media = (LPFLOPPYDISK)calloc(1, sizeof(FLOPPYDISK));
                if (media != NULL) {
                    media->image_raw_data = (LPVOID)malloc(siz);
                    if (media->image_raw_data != NULL) {
                        size_t len = fread(media->image_raw_data, 1, siz, fp);
                        if (len == siz) {
                            if (n80device.fdu.media_init(&n80device.fdu, media, media->image_raw_data, (DWORD)siz, d88_image_release)) {
                                // eject and insert floppy to drive
                                LPFDUDRIVE drv = &n80device.fdu.fdd.drives[unit];
                                drv->eject(drv);
                                result = drv->insert(drv, media);
                            }
                        }
                    }
                }
                if (!result)
                    d88_image_release(media);
            }
            fclose(fp);
        }
    }
    [self leavePause:false];
    return result;
}


#if TARGET_OS_IPHONE
+(bytecoder*)loadFile:(NSString*)fullpath {
	FILE *fp = NULL;
	fp = fopen([fullpath fileSystemRepresentation], "r");
	if (fp == NULL)
		return nil;
	fpos_t fsize = 0;
	fseek(fp, 0, SEEK_END);
	fgetpos(fp, &fsize);
	fseek(fp, 0, SEEK_SET);
	UInt8 buff[fsize];
	size_t len = fread(buff, 1, (size_t)fsize, fp);
	fclose(fp);
	
	bytecoder *data = [[bytecoder alloc] init];
	[data alloc_buffer:(int)len];
	[data push:buff Length:(int)len];
	return data;
}
#endif
//-----------------------------------------------------------------------------
//    Z80 CPU Thread
//-----------------------------------------------------------------------------
-(void)enter_z80_thread {
    volatile double est_time = system_tick_time() + IVLSEC;
    for (unsigned int tk = 0;; ++tk) {
        int expected = 1;
        if (atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 9)) {
            NSLog(@"Accept Request CPU HALT");
            self.running = false;
            atomic_store(&self->request_cpu_signal, 0);
            return;
        }
        expected = 2;
        if (atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 9)) {
            printf("Accept Request CPU RESET\n");
            [self cpu_reset];
            self->state = 0;
            est_time = system_tick_time() + IVLSEC;
            atomic_store(&self->request_cpu_signal, 0);
            continue;
        }
        expected = 3;
        if (atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 9)) {
            NSLog(@"Accept Request CPU PAUSE");
            self.pause = true;
            [pauseZ80sync lock];
            [pauseZ80sync unlock];
            self.pause = false;
            self->state = 0;
            n80device.dmac._penarty = 0;
            est_time = system_tick_time() + IVLSEC;
            atomic_store(&self->request_cpu_signal, 0);
            continue;
        }
        // wait for CPU HALT request
        if (self.terminating || n80device.cpu.r.x.HALT != 0)
            continue;
        
		// SUB CPU命令実行
		n80device.fdu.cpu.step(&n80device.fdu.cpu);

        // MAIN CPU命令実行
        if (n80device.dmac._penarty > 0) {
            if (n80device.dmac._penarty > 2) {
                n80device.cpu.r.x.T = 2;
                n80device.dmac._penarty -= 2;
            } else {
                n80device.cpu.r.x.T  = n80device.dmac._penarty;
                n80device.dmac._penarty = 0;
            }
            n80device.cpu.r.x.TwM1 = 0;
            n80device.cpu.r.x.W    = 0;
            n80device.cpu.r.x.Wh   = 0;
        } else {
            // Z80 step execute
            n80device.cpu.step(&n80device.cpu);
            if (n80device.cpu.halt_detect) {
                n80device.cpu.halt_detect = 0;
                // notify if detect illegal instructions
                if (self.halt_callback != nil) {
                    dispatch_async(dispatch_get_main_queue(), ^{
                        self.halt_callback(n80device.cpu.r.x.HALT);
                    });
                }
                [self req_reset];
            }
        }
        n80device.cpu.r.x.T += (n80device.cpu.r.x.TwM1 + n80device.cpu.r.x.W + n80device.cpu.r.x.Wh);
        
        // vrtc operation
        int vrtc = n80device.vrtc.blanking(&n80device.cpu);
        if (vrtc == _VRTC_VBBGN) {
            ;
        } else if (vrtc == _VRTC_VBEND) {
            // draw screen
            [self.n80gr draw_offscreen];
        }
        // peripheralls
        n80device.dispatch(&n80device);
        
        // beep/pcg ticks
        [self->sound append_sound_ticks:n80device.cpu.r.x.T];
        
        // cpu clock sync
        self->state += n80device.cpu.r.x.T;
        if (self->state >= n80device.param.sync_ticks) {
            while (1) {
                double cur_time = system_tick_time();
                if (cur_time >= est_time) {
                    est_time += IVLSEC;
                    if (cur_time > est_time) {
                        NSLog(@"Main CPU overrun!");
                        est_time = cur_time + IVLSEC;
                        self->state = 0;
                    } else
                        self->state -= n80device.param.sync_ticks;
                    break;
                }
                double remain_tick = est_time - cur_time;
                if (remain_tick <= 0.001)
                    continue;
                struct timespec req;
                req.tv_sec = 0;
                req.tv_nsec = (remain_tick - 0.001) * 1e9;
                nanosleep(&req, NULL);
            }
        }
    }
}

-(void)cpu_reset {
    n80device.cpu.reset(&n80device.cpu);
    n80device.reset(&n80device);
    [self->sound cpu_reset];
}
//-----------------------------------------------------------------------------
//    Draw off-screen
//-----------------------------------------------------------------------------
-(void)refleshScreen {
    [self.n80gr draw_offscreen];
}
//-----------------------------------------------------------------------------
//    Key Event Handler
//-----------------------------------------------------------------------------
-(void)key_clear {
    [pauseZ80sync lock];
    int expected = 0;
    while (!atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 3)) {
        expected = 0;
    }
    //OSAtomicCompareAndSwap32(0, 3, &n80lx.request_disp_signal);
    
    n80device.io.clear_keys();
    
    [pauseZ80sync unlock];
}

-(void)enterPause {
    if (!self.pause) {
        [pauseZ80sync lock];
        int expected = 0;
        while (!atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 3)) {
            expected = 0;
        }
        //OSAtomicCompareAndSwap32(0, 3, &n80lx.request_disp_signal);
    }
}

-(void)leavePause:(BOOL)terminating {
    if (self.pause) {
        [pauseZ80sync unlock];
    }
    self.terminating = terminating;
}

@end

