//
//  n80em.m
//
//  Copyright (C) 1993-2019 Hiroshi Kanemoto.
//    Released under the MIT license
//    http://opensource.org/licenses/mit-license.php
//
//     Created:    2012/08/22    H.Kanemoto
//    Modified:    2016/04/05    H.Kanemoto    ported from MaxOS X
//    Modified:    2019/05/12    H.Kanemoto    support METAL
//
#import "n80em.h"
#import "n80graph.h"
//-----------------------------------------------------------------------------
//    Emulator C Functions
//-----------------------------------------------------------------------------
extern N80LX        n80lx;
static const DWORD  SPITZ_COLOR[8]  = {0xFF000000,0xFFff0000,0xFF00ff00,0xFFffff00,0xFF0000ff,0xFFff00ff,0xFF00ffff,0xFFffffff};
static const DWORD  SPITZ_COLOR2[8] = {0xFF000000,0xFF7f0000,0xFF007f00,0xFF7f7f00,0xFF00007f,0xFF7f007f,0xFF007f7f,0xFF7f7f7f};
static const int    pitch = 640;

static void         _blt4020(LPDWORD, int line);
static void         _blt4025(LPDWORD, int line);
static void         _blt8020(LPDWORD, int line);
static void         _blt8025(LPDWORD, int line);
static void         bltblank(LPDWORD);

typedef void        (*BLTFNC)(LPDWORD, int);
static BLTFNC       bltfncs[] = {_blt4020,_blt4025,_blt8020,_blt8025};
static BLTFNC       fnc = _blt4020;
static DWORD        cur_pix_rgb, cur_pix_rgb2;
static int          vtbl_vram20[20], vtbl_vram25[25];
static LPCDWORD     second_rgb_colors = SPITZ_COLOR2;
//-----------------------------------------------------------------------------
//    Emulator Objective-C Class
//-----------------------------------------------------------------------------
@implementation n80graph

-(id)initWithN80:(n80em*)n80em {
    if (self = [super init]) {
        self->n80em = n80em;
        
        // lock for n80 vram sync
        self->sync = [[NSRecursiveLock alloc] init];
        [self prepareDotPatterns];
        self->draw_queue = dispatch_queue_create("n80_draw_queue", NULL);
    }
    return self;
}

-(void)dealloc {
    NSLog(@"n80graph dealloc");
    [self reqDispose];
}

-(void)lock {
    [self->sync lock];
}

-(void)unlock {
    [self->sync unlock];
}

-(void)suspend {
    atomic_store(&self->drawsync, 0);
}

-(void)resume {
}

- (void)start {
    atomic_store(&self->drawsync, 0);
}
- (void)reqDispose {
    //dispatch_release(self->draw_queue);
    self->sync = nil;
}
//-----------------------------------------------------------------------------
//    Initialize Emulator
//-----------------------------------------------------------------------------
-(void)prepareDotPatterns {
    // fb用ビットパターンデータを作成
    for (int line = 0; line < 20; ++line)
        vtbl_vram20[line] = pitch * (399 - (line * 20));
    for (int line = 0; line < 25; ++line)
        vtbl_vram25[line] = pitch * (399 - (line * 16));
}
-(void)chg_bltfunc:(int)mode {
    // mode: 0 = 40x20, 1 = 40x25, 2 = 80x20, 3 = 80x25
    fnc = bltfncs[mode];
    self->bltmode = mode;
}
-(void)chg_palette:(bool)interrace {
    second_rgb_colors = interrace ? SPITZ_COLOR2 : SPITZ_COLOR;
}
//-----------------------------------------------------------------------------
//    Draw off-screen
//-----------------------------------------------------------------------------
-(void)draw_offscreen {
    if (self->n80em.running) {
        int expected = 0;
        if (atomic_compare_exchange_strong(&self->drawsync, &expected, 1)) {
            dispatch_async(self->draw_queue, ^{
                [self draw_offscreen_core];
                atomic_store(&self->drawsync, 0);
            });
            //[self draw_offscreen_core];
            //atomic_store(&self->drawsync, 0);
        }
    }
}
-(void)draw_offscreen_core {
    // purchase bitmap pointer to off-screen
    self->bitmap = (LPDWORD)self->n80em.bmpimg.bitmap_back;
    
    if ((n80device.crtc._text_visible & 0x06) == 0) {
        for (int i = 0; i < 25; ++i) {
            n80device.vram_sync_enter();
            n80device.attr.cache(i);
            n80device.vram_sync_leave();
            n80device.attr.eval(i);
            fnc(self->bitmap, i);
        }
    } else {
        bltblank(self->bitmap);
    }
    //exchange off-screen and primary-screen
    [self->n80em.bmpimg flip];
}
-(void)draw_config:(char[25][80])text attr:(char[25][80])attr {
    // purchase bitmap pointer to off-screen
    self->bitmap = (LPDWORD)self->n80em.bmpimg.bitmap_back;

	for (int r = 0; r < 25; ++r) {
		LPBYTE txts = &n80device.attr._textcach[0];
		LPBYTE pacs = &n80device.attr._eval_attr[0];
		n80device.vram_sync_enter();
		for (int c = 0; c < 80; ++c) {
			*txts++ = (BYTE)text[r][c];
			*pacs++ = (BYTE)attr[r][c];
		}
		n80device.vram_sync_leave();
        fnc(self->bitmap, r);
	}
    [self draw_icon:12 ypos:16];
    //exchange off-screen and primary-screen
    [self->n80em.bmpimg flip];
    atomic_store(&self->drawsync, 0);
}
//---------------------------------------------------------------------------
//        draw off-screen bitmap
//---------------------------------------------------------------------------
static inline void set_pixel_80(LPDWORD vram) {
    *vram           = cur_pix_rgb;
    *(vram-pitch)   = cur_pix_rgb2;
}
static inline void set_pixel_40(LPDWORD vram) {
    *vram           = cur_pix_rgb;
    *(vram-pitch)   = cur_pix_rgb2;
    *(vram+1)       = cur_pix_rgb;
    *(vram-pitch+1) = cur_pix_rgb2;
}
static inline void reset_pixel_80(LPDWORD vram) {
    *vram           = 0;
    *(vram-pitch)   = 0;
}
static inline void reset_pixel_40(LPDWORD vram) {
    *vram           = 0;
    *(vram-pitch)   = 0;
    *(vram+1)       = 0;
    *(vram-pitch+1) = 0;
}
//---------------------------------------------------------------------------
//        process font pattern by attribute
//---------------------------------------------------------------------------
static inline BYTE proc_video_pixcel25(BYTE attr, BYTE bit, int r) {
	if (attr & _PAC_ATTR_UND_LINE && r == 7)	// under line
        bit |= 0xff;
	if (attr & _PAC_ATTR_UPR_LINE && r == 0)	// upper line
        bit |= 0xff;
	if (attr & _PAC_ATTR_ERASE)					// secret
        bit = 0;
	if (attr & _PAC_ATTR_REVERSE)				// reverse
        bit = ~bit;
    return bit;
}
static inline BYTE proc_video_pixcel20(BYTE attr, BYTE bit, int r) {
	if (attr & _PAC_ATTR_UND_LINE && r == 9)	// under line
        bit |= 0xff;
	if (attr & _PAC_ATTR_UPR_LINE && r == 0)	// upper line
        bit |= 0xff;
	if (attr & _PAC_ATTR_ERASE)					// secret
        bit = 0;
	if (attr & _PAC_ATTR_REVERSE)				// reverse
        bit = ~bit;
    return bit;
}
static inline LPBYTE confirm_fnt_ptr(LPBYTE fnt, BYTE attr, int vsp) {
	if (n80device.param.use_new_pcg) {
		// attribute = REVERSE and new_pcg = ON
		if (attr & _PAC_ATTR_REVERSE && n80device.chiqlappe._swich == 1) {
			fnt += 4096;
			// new_pcg block mode
			if (n80device.chiqlappe._block_mode)
				fnt += (2048 * (vsp / 64));
		} else if (attr & _PAC_ATTR_GRAPH)
			fnt += 2048;
	} else {
		if (attr & _PAC_ATTR_GRAPH)		// graph
			fnt += 2048;
		else if (n80device.pcg_mode)	// pcg
			fnt += 4096;
	}
	return fnt;
}
//---------------------------------------------------------------------------
//        draw Text(40x20)
//---------------------------------------------------------------------------
static inline LPDWORD draw_font_raster4020(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
    register LPDWORD _vram = (vram - (pitch * 2));
    register BYTE pat = proc_video_pixcel20(attr, fnt, r);

    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);

    return _vram;
}
static void _blt4020(LPDWORD bitmap, int line) {
    if (line < 20) {
        LPBYTE  txts = &n80device.attr._textcach[0];
        LPBYTE  pacs = &n80device.attr._eval_attr[0];
        int     col, offset = vtbl_vram20[line];//pitch * (399 - (line * 20));

        for (col = 0; col < 80; col += 2) {
            BYTE        ascii = *txts;
            BYTE        attr  = *pacs;
            LPBYTE      fnt   = &n80device.font[ascii*8];
            LPDWORD     vram  = &bitmap[offset + (col * 8)];
            int         color = attr & 0x07;

            cur_pix_rgb  = SPITZ_COLOR[color];
            cur_pix_rgb2 = second_rgb_colors[color];
            
            txts += 2;
            pacs += 2;
            
			fnt = confirm_fnt_ptr(fnt, attr, line * 10);

            vram = draw_font_raster4020(vram, 0, attr, 0);
            vram = draw_font_raster4020(vram, 1, attr, *fnt++);
            vram = draw_font_raster4020(vram, 2, attr, *fnt++);
            vram = draw_font_raster4020(vram, 3, attr, *fnt++);
            vram = draw_font_raster4020(vram, 4, attr, *fnt++);
            vram = draw_font_raster4020(vram, 5, attr, *fnt++);
            vram = draw_font_raster4020(vram, 6, attr, *fnt++);
            vram = draw_font_raster4020(vram, 7, attr, *fnt++);
            vram = draw_font_raster4020(vram, 8, attr, *fnt++);
            vram = draw_font_raster4020(vram, 9, attr, 0);
        }
    }
}
//---------------------------------------------------------------------------
//        draw Text(40x25)
//---------------------------------------------------------------------------
static inline LPDWORD draw_font_raster4025(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
    register LPDWORD _vram = (vram - (pitch * 2));
    register BYTE pat = proc_video_pixcel25(attr, fnt, r);

    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);

    return _vram;
}
static void _blt4025(LPDWORD bitmap, int line) {
    if (line < 25) {
        LPBYTE  txts = &n80device.attr._textcach[0];
        LPBYTE  pacs = &n80device.attr._eval_attr[0];
        int     col, offset = vtbl_vram25[line];//pitch * (399 - (line * 16));
        
        for (col = 0; col < 80; col += 2) {
            BYTE        ascii = *txts;
            BYTE        attr  = *pacs;
            LPBYTE      fnt   = &n80device.font[ascii*8];
            LPDWORD     vram  = &bitmap[offset + (col * 8)];
            int         color = attr & 0x07;

            cur_pix_rgb  = SPITZ_COLOR[color];
            cur_pix_rgb2 = second_rgb_colors[color];
            
            txts += 2;
            pacs += 2;
            
			fnt = confirm_fnt_ptr(fnt, attr, line * 8);

            vram = draw_font_raster4025(vram, 0, attr, *fnt++);
            vram = draw_font_raster4025(vram, 1, attr, *fnt++);
            vram = draw_font_raster4025(vram, 2, attr, *fnt++);
            vram = draw_font_raster4025(vram, 3, attr, *fnt++);
            vram = draw_font_raster4025(vram, 4, attr, *fnt++);
            vram = draw_font_raster4025(vram, 5, attr, *fnt++);
            vram = draw_font_raster4025(vram, 6, attr, *fnt++);
            vram = draw_font_raster4025(vram, 7, attr, *fnt++);
        }
    }
}
//---------------------------------------------------------------------------
//        draw Text(80x20)
//---------------------------------------------------------------------------
static inline LPDWORD draw_font_raster8020(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
    register LPDWORD _vram = (vram - (pitch * 2));
    register BYTE pat = proc_video_pixcel20(attr, fnt, r);

    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram) : reset_pixel_80(vram);

    return _vram;
}
static void _blt8020(LPDWORD bitmap, int line) {
    if (line < 20) {
        LPBYTE  txts = &n80device.attr._textcach[0];
        LPBYTE  pacs = &n80device.attr._eval_attr[0];
        int     col, offset = vtbl_vram20[line];//pitch * (399 - (line * 20));
        
        for (col = 0; col < 80; ++col) {
            BYTE        ascii = *txts++;
            BYTE        attr  = *pacs++;
            LPBYTE      fnt   = &n80device.font[ascii*8];
            LPDWORD     vram  = &bitmap[offset + (col * 8)];
            int         color = attr & 0x07;

            cur_pix_rgb  = SPITZ_COLOR[color];
            cur_pix_rgb2 = second_rgb_colors[color];
            
			fnt = confirm_fnt_ptr(fnt, attr, line * 10);

            vram = draw_font_raster8020(vram, 0, attr, 0);
            vram = draw_font_raster8020(vram, 1, attr, *fnt++);
            vram = draw_font_raster8020(vram, 2, attr, *fnt++);
            vram = draw_font_raster8020(vram, 3, attr, *fnt++);
            vram = draw_font_raster8020(vram, 4, attr, *fnt++);
            vram = draw_font_raster8020(vram, 5, attr, *fnt++);
            vram = draw_font_raster8020(vram, 6, attr, *fnt++);
            vram = draw_font_raster8020(vram, 7, attr, *fnt++);
            vram = draw_font_raster8020(vram, 8, attr, *fnt++);
            vram = draw_font_raster8020(vram, 9, attr, 0);
        }
    }
}
//---------------------------------------------------------------------------
//        draw Text(80x25)
//---------------------------------------------------------------------------
static inline LPDWORD draw_font_raster8025(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
    register LPDWORD _vram = (vram - (pitch * 2));
    register BYTE pat = proc_video_pixcel25(attr, fnt, r);

    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram) : reset_pixel_80(vram);

    return _vram;
}
static void _blt8025(LPDWORD bitmap, int line) {
    if (line < 25) {
        LPBYTE  txts = &n80device.attr._textcach[0];
        LPBYTE  pacs = &n80device.attr._eval_attr[0];
        int     col, offset = vtbl_vram25[line];//pitch * (399 - (line * 16));

        for (col = 0; col < 80; ++col) {
            BYTE        ascii = *txts++;
            BYTE        attr  = *pacs++;
            LPBYTE      fnt   = &n80device.font[ascii*8];
            LPDWORD     vram  = &bitmap[offset + (col * 8)];
            int         color = attr & 0x07;

            cur_pix_rgb  = SPITZ_COLOR[color];
            cur_pix_rgb2 = second_rgb_colors[color];
            
			fnt = confirm_fnt_ptr(fnt, attr, line * 8);

            vram = draw_font_raster8025(vram, 0, attr, *fnt++);
            vram = draw_font_raster8025(vram, 1, attr, *fnt++);
            vram = draw_font_raster8025(vram, 2, attr, *fnt++);
            vram = draw_font_raster8025(vram, 3, attr, *fnt++);
            vram = draw_font_raster8025(vram, 4, attr, *fnt++);
            vram = draw_font_raster8025(vram, 5, attr, *fnt++);
            vram = draw_font_raster8025(vram, 6, attr, *fnt++);
            vram = draw_font_raster8025(vram, 7, attr, *fnt++);
        }
    }
}
//---------------------------------------------------------------------------
//        clear off-screen bitmap
//---------------------------------------------------------------------------
static void bltblank(LPDWORD bitmap) {
    int h = 0, i = 0;
    for (h = 0; h < 20; ++h) {
        int offset = vtbl_vram20[h];
        LPDWORD    vram = &bitmap[offset];
        for (i = 0; i < 640; ++i) {
            vram[i]          = 0;
            vram[i-pitch]    = 0;
            vram[i-pitch*2]  = 0;
            vram[i-pitch*3]  = 0;
            vram[i-pitch*4]  = 0;
            vram[i-pitch*5]  = 0;
            vram[i-pitch*6]  = 0;
            vram[i-pitch*7]  = 0;
            vram[i-pitch*8]  = 0;
            vram[i-pitch*9]  = 0;
            vram[i-pitch*10] = 0;
            vram[i-pitch*11] = 0;
            vram[i-pitch*12] = 0;
            vram[i-pitch*13] = 0;
            vram[i-pitch*14] = 0;
            vram[i-pitch*15] = 0;
            vram[i-pitch*16] = 0;
            vram[i-pitch*17] = 0;
            vram[i-pitch*18] = 0;
            vram[i-pitch*19] = 0;
        }
    }
}
//---------------------------------------------------------------------------
//        ICON data
//---------------------------------------------------------------------------
typedef struct rgb {
	BYTE	r;
	BYTE	g;
	BYTE	b;
}	rgb;

static rgb	icon[32][32] = {

{{0x0,0x2B,0x3D},{0x0,0x7F,0xB6},{0x0,0xAE,0xF9},{0x0,0xA7,0xEF},{0x0,0x9F,0xE4},{0x0,0xA1,0xE7},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE9},{0x0,0x9F,0xE3},{0x0,0xAA,0xF4},{0x0,0x90,0xCE},{0x0,0x1A,0x25}},
{{0x0,0x0,0x0},{0x0,0x3,0x5},{0x0,0x41,0x5D},{0x0,0x8F,0xCD},{0x0,0xAE,0xF8},{0x0,0xA4,0xEC},{0x0,0x9F,0xE4},{0x0,0xA2,0xE7},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0x9F,0xE4},{0x0,0xA9,0xF2},{0x0,0x95,0xD5},{0x0,0x1A,0x25},{0x0,0x0,0x0}},
{{0x0,0x1,0x2},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x7,0xA},{0x0,0x4E,0x6F},{0x0,0x99,0xDB},{0x0,0xAE,0xF9},{0x0,0xA3,0xE9},{0x0,0x9F,0xE4},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0x9F,0xE4},{0x0,0xAA,0xF3},{0x0,0x91,0xCF},{0x0,0x17,0x21},{0x0,0x0,0x0},{0x0,0x2,0x3}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x2,0x3},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x11,0x18},{0x0,0x5E,0x85},{0x0,0xA1,0xE6},{0x0,0xAC,0xF7},{0x0,0xA1,0xE7},{0x0,0xA0,0xE5},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0x9F,0xE3},{0x0,0xAA,0xF4},{0x0,0x90,0xCE},{0x0,0x14,0x1B},{0x0,0x0,0x0},{0x0,0x2,0x3},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x1},{0x0,0x2,0x3},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x1B,0x27},{0x0,0x6D,0x9B},{0x1,0xA7,0xEF},{0x3,0xAE,0xF4},{0x4,0xA4,0xE5},{0x4,0xA3,0xE4},{0x4,0xA3,0xE5},{0x2,0xA2,0xE7},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x1,0xA2,0xE7},{0x3,0xA3,0xE6},{0x3,0xA1,0xE2},{0x2,0xAD,0xF4},{0x0,0x8D,0xCA},{0x0,0x13,0x1B},{0x0,0x0,0x0},{0x0,0x2,0x3},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x1,0x1},{0x0,0x2,0x3},{0x0,0x0,0x0},{0x2,0x0,0x0},{0x0,0x26,0x39},{0x0,0x6C,0xAF},{0x0,0x96,0xF5},{0x0,0x98,0xF9},{0x0,0x9A,0xF7},{0x0,0xA0,0xED},{0x3,0xA3,0xE5},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x3,0xA3,0xE5},{0x0,0xA2,0xE9},{0x0,0x9B,0xF1},{0x0,0x9E,0xFD},{0x0,0x82,0xC9},{0x2,0x12,0x18},{0x1,0x0,0x0},{0x0,0x2,0x3},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x2,0x3,0x2},{0x0,0x0,0x3},{0x4,0x0,0x0},{0x46,0x43,0x0},{0x82,0xB4,0x4A},{0x8F,0xE2,0x7C},{0x62,0xC7,0x93},{0x1A,0xAA,0xD3},{0x0,0x9D,0xF5},{0x2,0xA3,0xE6},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x3,0xA3,0xE5},{0x0,0x9D,0xF4},{0x4,0xA1,0xE6},{0x3E,0xBC,0xB7},{0x5D,0xC8,0x9C},{0x2F,0x3F,0x17},{0x0,0x0,0x0},{0x0,0x2,0x3},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0x0,0x0,0x0},{0x31,0x31,0x0},{0xCF,0xD0,0x2},{0xFF,0xFF,0x1},{0xB4,0xAA,0x0},{0x72,0x66,0x0},{0xF2,0xE9,0x0},{0xEE,0xF8,0x10},{0x61,0xC2,0x93},{0x0,0x9D,0xF4},{0x2,0xA1,0xE2},{0x0,0xA1,0xE8},{0x4,0xA3,0xE4},{0x0,0x9C,0xF7},{0x4B,0xBA,0xA7},{0xD7,0xF0,0x28},{0xEA,0xE7,0x0},{0xB1,0xA7,0x0},{0xF1,0xEE,0x0},{0xA2,0xA3,0x1},{0x5,0x5,0x0},{0x2,0x2,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0x1,0x1,0x0},{0x1D,0x1D,0x0},{0xEA,0xEA,0x0},{0xFF,0xFF,0x0},{0xDA,0xDA,0x0},{0x0,0x1,0x3},{0x0,0x0,0x4},{0x60,0x63,0x4},{0xFF,0xFF,0x0},{0xFF,0xFE,0x0},{0x3F,0xB8,0xB1},{0x0,0xA7,0xFF},{0x6,0xA6,0xE3},{0x0,0x9B,0xF7},{0x5B,0xC1,0x98},{0xFF,0xFE,0x0},{0xF9,0xF5,0x0},{0x28,0x2A,0x2},{0x0,0x0,0x3},{0x61,0x62,0x1},{0xFF,0xFF,0x0},{0x89,0x89,0x0},{0x0,0x0,0x0},{0x3,0x3,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x4,0x4,0x0},{0x0,0x0,0x0},{0x96,0x96,0x0},{0xFF,0xFF,0x0},{0xFD,0xFD,0x0},{0xA1,0xA1,0x0},{0x0,0x0,0x0},{0x4,0x4,0x0},{0x40,0x40,0x0},{0xFE,0xFE,0x2},{0xFF,0xFF,0x2},{0xAD,0xB7,0x14},{0x0,0x56,0x8C},{0x0,0xA0,0xF1},{0x3A,0xB7,0xB8},{0xFB,0xFC,0x2},{0xFF,0xFF,0x2},{0x97,0x99,0x3},{0x0,0x0,0x0},{0x3,0x3,0x0},{0x3E,0x3E,0x0},{0xFC,0xFC,0x0},{0xF3,0xF3,0x0},{0x16,0x16,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0xD8,0xD8,0x0},{0xFE,0xFE,0x0},{0xFF,0xFF,0x0},{0x92,0x92,0x0},{0x2,0x2,0x0},{0x0,0x0,0x0},{0x44,0x44,0x0},{0xFF,0xFF,0x0},{0xFE,0xFE,0x1},{0xC9,0xC7,0x0},{0x1,0x0,0x0},{0xD,0x2A,0x32},{0xDE,0xF1,0x1C},{0xFF,0xFF,0x0},{0xFF,0xFF,0x2},{0x3F,0x3F,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0},{0x61,0x61,0x0},{0xFF,0xFF,0x0},{0xFF,0xFF,0x0},{0x4E,0x4E,0x0},{0x0,0x0,0x0},{0x3,0x3,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0},{0x8,0x8,0x0},{0xE3,0xE3,0x0},{0xFF,0xFF,0x0},{0xFF,0xFF,0x0},{0xAF,0xAF,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x58,0x58,0x0},{0xFE,0xFE,0x0},{0xFE,0xFE,0x0},{0xB3,0xB4,0x1},{0x0,0x0,0x3},{0x89,0x84,0x0},{0xFF,0xFF,0x0},{0xFE,0xFF,0x1},{0xDF,0xDF,0x0},{0x6,0x6,0x0},{0x3,0x3,0x0},{0x0,0x0,0x0},{0x8E,0x8E,0x0},{0xFF,0xFF,0x0},{0xFE,0xFE,0x0},{0x70,0x70,0x0},{0x0,0x0,0x0},{0x4,0x4,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x1,0x1,0x0},{0xCE,0xCE,0x0},{0xFE,0xFE,0x0},{0xFD,0xFD,0x0},{0xF0,0xF0,0x0},{0x1B,0x1B,0x0},{0x0,0x0,0x0},{0x84,0x84,0x0},{0xFE,0xFE,0x0},{0xFF,0xFF,0x0},{0x41,0x41,0x0},{0x12,0x12,0x0},{0xF5,0xF7,0x2},{0xFB,0xFC,0x1},{0xFE,0xFE,0x0},{0x95,0x95,0x0},{0x0,0x0,0x0},{0x5,0x5,0x0},{0x0,0x0,0x0},{0xC7,0xC7,0x0},{0xFF,0xFF,0x0},{0xFF,0xFF,0x0},{0x75,0x75,0x0},{0x0,0x0,0x0},{0x4,0x4,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x4,0x4,0x0},{0x0,0x0,0x0},{0x85,0x85,0x0},{0xFF,0xFF,0x0},{0xF8,0xF8,0x0},{0xFF,0xFF,0x0},{0xA0,0xA0,0x0},{0x10,0x10,0x0},{0xEB,0xEB,0x0},{0xFF,0xFF,0x0},{0x79,0x79,0x0},{0x0,0x0,0x0},{0x7F,0x7F,0x0},{0xFF,0xFF,0x0},{0xF9,0xF9,0x0},{0xFF,0xFF,0x0},{0x47,0x47,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x1C,0x1C,0x0},{0xF6,0xF6,0x0},{0xFD,0xFD,0x0},{0xFF,0xFF,0x0},{0x66,0x66,0x0},{0x0,0x0,0x0},{0x4,0x4,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0x2,0x2,0x0},{0x1E,0x1E,0x0},{0xEB,0xEB,0x0},{0xFB,0xFB,0x0},{0xFD,0xFD,0x0},{0xFD,0xFD,0x0},{0xD7,0xD7,0x0},{0xB4,0xB4,0x0},{0x45,0x45,0x0},{0x0,0x0,0x0},{0x7,0x7,0x0},{0xDA,0xDA,0x0},{0xFF,0xFF,0x0},{0xFF,0xFF,0x0},{0xE9,0xE9,0x0},{0xC,0xC,0x0},{0x2,0x2,0x0},{0x0,0x0,0x0},{0x5A,0x5A,0x0},{0xFF,0xFF,0x0},{0xF8,0xF8,0x0},{0xFF,0xFF,0x0},{0x4C,0x4C,0x0},{0x0,0x0,0x0},{0x3,0x3,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x86,0x86,0x0},{0xFF,0xFF,0x0},{0xF9,0xF9,0x0},{0xFD,0xFD,0x0},{0xFF,0xFF,0x0},{0x37,0x37,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0x38,0x38,0x0},{0xFF,0xFF,0x0},{0xFA,0xFA,0x0},{0xFF,0xFF,0x0},{0xA9,0xA9,0x0},{0x0,0x0,0x0},{0x7,0x7,0x0},{0x0,0x0,0x0},{0xA2,0xA2,0x0},{0xFF,0xFF,0x0},{0xFC,0xFC,0x0},{0xFD,0xFD,0x0},{0x28,0x28,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x3,0x3,0x0},{0x0,0x0,0x0},{0x1C,0x1C,0x0},{0xA2,0xA2,0x0},{0xD7,0xD7,0x0},{0xB2,0xB2,0x0},{0xFF,0xFF,0x0},{0xFB,0xFB,0x0},{0xFE,0xFE,0x0},{0xE5,0xE5,0x0},{0x17,0x17,0x0},{0x1,0x1,0x0},{0x7D,0x7D,0x0},{0xFF,0xFF,0x0},{0xF7,0xF7,0x0},{0xFF,0xFF,0x0},{0x5F,0x5F,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0x8,0x8,0x0},{0xE3,0xE3,0x0},{0xFF,0xFF,0x0},{0xFF,0xFF,0x0},{0xDF,0xDF,0x0},{0x6,0x6,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0x0,0x0,0x0},{0x38,0x38,0x0},{0xE5,0xE5,0x0},{0xFF,0xFF,0x0},{0x50,0x50,0x0},{0x0,0x0,0x0},{0xC6,0xC6,0x0},{0xFF,0xFF,0x0},{0xF9,0xF9,0x0},{0xFF,0xFF,0x0},{0x9A,0x9A,0x0},{0x0,0x0,0x0},{0xB4,0xB4,0x0},{0xFF,0xFF,0x0},{0xFD,0xFD,0x0},{0xF8,0xF8,0x0},{0x1E,0x1E,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x3C,0x3C,0x0},{0xFF,0xFF,0x0},{0xFA,0xFA,0x0},{0xFE,0xFE,0x0},{0xA4,0xA4,0x0},{0x0,0x0,0x0},{0x3,0x3,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0x1,0x1,0x0},{0x1B,0x1B,0x0},{0xE9,0xE9,0x0},{0xFF,0xFF,0x0},{0xC7,0xC7,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x23,0x23,0x0},{0xF4,0xF4,0x0},{0xFD,0xFD,0x0},{0xFD,0xFD,0x0},{0xF5,0xF5,0x0},{0x15,0x15,0x0},{0xCE,0xCE,0x0},{0xFF,0xFF,0x0},{0xFF,0xFF,0x0},{0xC8,0xC8,0x0},{0x0,0x0,0x0},{0x5,0x5,0x0},{0x0,0x0,0x0},{0x85,0x85,0x0},{0xFF,0xFF,0x0},{0xF8,0xF8,0x0},{0xFF,0xFF,0x0},{0x52,0x52,0x0},{0x0,0x0,0x0},{0x4,0x4,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x4,0x4,0x0},{0x0,0x0,0x0},{0x93,0x93,0x0},{0xFF,0xFF,0x0},{0xFD,0xFD,0x0},{0x7C,0x7C,0x0},{0x0,0x0,0x0},{0x8,0x8,0x0},{0x0,0x0,0x0},{0xA3,0xA3,0x0},{0xFF,0xFF,0x0},{0xFB,0xFB,0x0},{0xFF,0xFF,0x0},{0x47,0x47,0x0},{0xDB,0xDB,0x0},{0xFF,0xFF,0x0},{0xFF,0xFF,0x0},{0x85,0x85,0x0},{0x0,0x0,0x0},{0x5,0x5,0x0},{0x1,0x1,0x0},{0xCD,0xCD,0x0},{0xFF,0xFF,0x0},{0xFD,0xFD,0x0},{0xE5,0xE5,0x0},{0xB,0xB,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x2,0x2,0x0},{0xD7,0xD7,0x0},{0xFE,0xFE,0x0},{0xFF,0xFF,0x0},{0x45,0x45,0x0},{0x0,0x0,0x0},{0x7,0x7,0x0},{0x1,0x1,0x0},{0x68,0x68,0x0},{0xFF,0xFF,0x0},{0xFB,0xFB,0x0},{0xFE,0xFF,0x2},{0x4A,0x4C,0x3},{0xE2,0xE2,0x0},{0xFF,0xFF,0x0},{0xFF,0xFF,0x0},{0x43,0x43,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x27,0x27,0x0},{0xFB,0xFB,0x0},{0xFB,0xFB,0x0},{0xFF,0xFF,0x0},{0x7A,0x7A,0x0},{0x0,0x0,0x0},{0x4,0x4,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0},{0xA,0xA,0x0},{0xE5,0xE5,0x0},{0xFF,0xFF,0x0},{0xFE,0xFE,0x0},{0x2A,0x2A,0x0},{0x0,0x0,0x0},{0x6,0x6,0x0},{0x1,0x1,0x0},{0x64,0x64,0x0},{0xFF,0xFF,0x0},{0xFD,0xFE,0x1},{0xEF,0xE9,0x0},{0x21,0x1B,0x0},{0xE3,0xE4,0x1},{0xFF,0xFF,0x0},{0xEF,0xEF,0x0},{0x13,0x13,0x0},{0x1,0x1,0x0},{0x1,0x1,0x0},{0x70,0x70,0x0},{0xFF,0xFF,0x0},{0xFC,0xFC,0x0},{0xE5,0xE5,0x0},{0xF,0xF,0x0},{0x1,0x1,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x1,0x1,0x0},{0xD1,0xD1,0x0},{0xFD,0xFD,0x0},{0xFA,0xFA,0x0},{0x24,0x24,0x0},{0x2,0x2,0x0},{0x8,0x8,0x0},{0x0,0x0,0x0},{0x93,0x93,0x0},{0xFD,0xFE,0x4},{0xFF,0xFF,0x0},{0xA9,0xD4,0x3F},{0x0,0x23,0x4C},{0xCC,0xC5,0x0},{0xFE,0xFE,0x2},{0xCF,0xCF,0x0},{0x1,0x2,0x0},{0x2,0x2,0x0},{0x0,0x0,0x0},{0xBF,0xBF,0x0},{0xFC,0xFC,0x0},{0xFF,0xFF,0x0},{0x55,0x55,0x0},{0x0,0x0,0x0},{0x3,0x3,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x4,0x4,0x0},{0x0,0x0,0x0},{0x71,0x71,0x0},{0xFF,0xFF,0x0},{0xFB,0xFB,0x0},{0x2F,0x2F,0x0},{0x0,0x0,0x2},{0x0,0x0,0x1},{0xF,0x12,0x5},{0xE4,0xE7,0x5},{0xFF,0xFF,0x0},{0xEC,0xF7,0x11},{0x23,0xB1,0xD3},{0x0,0x9B,0xFF},{0x7C,0xC2,0x69},{0xFF,0xFF,0x0},{0xB4,0xB8,0x6},{0x0,0x0,0x0},{0x0,0x0,0x1},{0x22,0x22,0x1},{0xF6,0xF6,0x0},{0xFF,0xFF,0x0},{0x94,0x94,0x0},{0x0,0x0,0x0},{0x3,0x3,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x2,0x2,0x0},{0x4,0x4,0x0},{0xA4,0xA4,0x0},{0xFF,0xFF,0x1},{0xC2,0xC4,0x2},{0x26,0x20,0x0},{0x2F,0x32,0x8},{0xCB,0xD0,0xB},{0xFF,0xFF,0x0},{0xDD,0xF2,0x20},{0x36,0xB3,0xBD},{0x0,0x9D,0xF1},{0x1,0xA2,0xE8},{0x11,0xA7,0xE0},{0xE3,0xF5,0x1B},{0xEF,0xE6,0x0},{0x14,0x19,0x8},{0xA,0x7,0x0},{0xC0,0xBF,0x0},{0xFF,0xFF,0x3},{0x99,0x9A,0x0},{0x1,0x1,0x0},{0x3,0x3,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x1,0x1,0x0},{0x0,0x0,0x0},{0x0,0x0,0x2},{0x5A,0x59,0x0},{0xB9,0xB3,0x0},{0xCA,0xE1,0x23},{0xD2,0xED,0x2A},{0xC5,0xEE,0x3E},{0x74,0xCA,0x81},{0x11,0xA6,0xDD},{0x0,0x9E,0xF2},{0x2,0xA3,0xE5},{0x2,0xA2,0xE5},{0x0,0x9E,0xF0},{0x27,0xAD,0xC9},{0xC0,0xE9,0x3F},{0xD7,0xE3,0x14},{0xCD,0xD7,0x12},{0xC9,0xCD,0x7},{0x54,0x4A,0x0},{0x0,0x0,0x1},{0x1,0x3,0x2},{0x1,0x1,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x3,0x4},{0x1,0x0,0x0},{0x0,0x0,0x7},{0x0,0x55,0x8F},{0x2,0xA6,0xED},{0x2,0xA2,0xE7},{0x0,0x9C,0xF1},{0x0,0x9B,0xF8},{0x0,0xA1,0xEB},{0x2,0xA3,0xE5},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x1,0xA3,0xE6},{0x0,0x9F,0xEF},{0x0,0x9E,0xEF},{0x20,0xAF,0xCF},{0x1E,0xB4,0xD8},{0x0,0x84,0xCB},{0x0,0x2E,0x56},{0x2,0x1,0x0},{0x0,0x0,0x0},{0x0,0x2,0x3},{0x0,0x1,0x2},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x1},{0x0,0x2,0x3},{0x0,0x0,0x0},{0x0,0x1B,0x26},{0x3,0x89,0xBF},{0x3,0xB1,0xF9},{0x0,0xA0,0xE7},{0x0,0xA1,0xE8},{0x2,0xA3,0xE6},{0x4,0xA3,0xE4},{0x1,0xA2,0xE7},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x2,0xA3,0xE6},{0x1,0xA3,0xE7},{0x0,0x9F,0xEE},{0x0,0x9D,0xEC},{0x1,0xA8,0xEF},{0x3,0xAF,0xF6},{0x0,0x80,0xB7},{0x0,0x2C,0x3E},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x2,0x3},{0x0,0x1,0x1},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x0,0x0}},
{{0x0,0x0,0x0},{0x0,0x1,0x2},{0x0,0x0,0x1},{0x0,0x0,0x0},{0x0,0x3B,0x54},{0x0,0x9E,0xE3},{0x0,0xAA,0xF3},{0x0,0x9F,0xE3},{0x0,0xA2,0xE8},{0x1,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x1,0xA2,0xE7},{0x1,0xA3,0xE7},{0x0,0xA1,0xE6},{0x0,0x9F,0xE5},{0x0,0xAA,0xF3},{0x0,0xA8,0xF1},{0x0,0x6F,0x9E},{0x0,0x1C,0x28},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x2,0x3},{0x0,0x0,0x1},{0x0,0x0,0x0}},
{{0x0,0x2,0x3},{0x0,0x0,0x0},{0x0,0x3,0x4},{0x0,0x60,0x89},{0x0,0xAC,0xF6},{0x0,0xA4,0xEB},{0x0,0xA0,0xE5},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA0,0xE5},{0x0,0xA1,0xE6},{0x0,0xAC,0xF7},{0x0,0xA1,0xE7},{0x0,0x5D,0x85},{0x0,0xE,0x15},{0x0,0x0,0x0},{0x0,0x0,0x0},{0x0,0x1,0x2}},
{{0x0,0x0,0x0},{0x0,0x1A,0x25},{0x0,0x83,0xBC},{0x0,0xAE,0xF9},{0x0,0xA0,0xE5},{0x0,0xA1,0xE7},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA0,0xE4},{0x0,0xA2,0xE9},{0x0,0xAD,0xF8},{0x0,0x98,0xD9},{0x0,0x4D,0x6E},{0x0,0x8,0xB},{0x0,0x0,0x0}},
{{0x0,0x34,0x4A},{0x0,0x99,0xDB},{0x0,0xAC,0xF6},{0x0,0x9F,0xE3},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE8},{0x0,0xA2,0xE7},{0x0,0x9F,0xE4},{0x0,0xA5,0xEC},{0x0,0xAF,0xFB},{0x0,0x88,0xC2},{0x0,0x33,0x49}},

};
//---------------------------------------------------------------------------
//        draw ICON in config
//---------------------------------------------------------------------------
-(void)draw_icon:(int)xpos ypos:(int)ypos {
    self->bitmap = (LPDWORD)self->n80em.bmpimg.bitmap_back;
    for (int y = 31; y >= 0; --y) {
        for (int x = 0; x < 32; ++ x) {
            [self->n80em.bmpimg setPixel:false X:xpos + x Y:(400 - ypos) - y R:icon[y][x].r G:icon[y][x].g B:icon[y][x].b];
        }
    }
}

@end

