//
//  n80em.m
//
//  Copyright (C) 1993-2019 Hiroshi Kanemoto.
//    Released under the MIT license
//    http://opensource.org/licenses/mit-license.php
//
//     Created:    2019/05/12    H.Kanemoto
//
#import "n80sound.h"
#import <stdatomic.h>
#include <limits.h>
#include <mach/mach_time.h>
#include <AudioUnit/AudioUnit.h>
extern N80LX        n80lx;
//-----------------------------------------------------------------------------
//    Emulator C Functions
//-----------------------------------------------------------------------------
#define SAMPLE_RATE (96000.0)

static SAMPLETsBANK    beep_bank;
static SAMPLETsBANK    sound_bank[3];
static SAMPLETsARR    beep_copy_arr;
static SAMPLETsARR    sound_copy_arr[3];
//-----------------------------------------------------------------------------
//    supplement function
//-----------------------------------------------------------------------------
static inline void append_tick_to_array(LPSAMPLETsBANK bank, DWORD new_state) {
    // 現在のTick累積を配列に追加する
    LPSAMPLETsARR arr = &bank->banks[bank->bank];
    if (arr->count < MAXSTsARRAYS - 1) {
        arr->arr[arr->count].ticks   = bank->current.ticks;
        arr->arr[arr->count++].state = bank->current.state;
        arr->ticks_sum += bank->current.ticks;
    } else
        fprintf(stderr, "beep ticks array overflow\n");
    
    bank->current.ticks = 0;
    bank->current.state = new_state;
}
static inline void advance_tick_to_array(LPSAMPLETsBANK bank, double diff) {
    double add = 0;
    if (bank->current.ticks > diff) {
        bank->current.ticks -= diff;
        add = diff;
    } else {
        add = bank->current.ticks;
        bank->current.ticks = 0;
    }
    if (add > 0) {
        LPSAMPLETsARR arr = &bank->banks[bank->bank];
        if (arr->count < MAXSTsARRAYS - 1) {
            arr->arr[arr->count].ticks = add;
            arr->arr[arr->count].state = bank->current.state;
            arr->ticks_sum += add;
            ++arr->count;
        } else if (arr->arr[MAXSTsARRAYS - 1].state == bank->current.state) {
            arr->arr[MAXSTsARRAYS - 1].ticks += add;
            arr->ticks_sum += add;
        } else
            fprintf(stderr, "advanceTick: bank overrun\n");
    }
}
//-----------------------------------------------------------------------------
//    AudioUnit Callback function
//-----------------------------------------------------------------------------
static OSStatus    audioCallback(
    void                           *inRefCon,
    AudioUnitRenderActionFlags     *ioActionFlags,
    const AudioTimeStamp           *inTimeStamp,
    UInt32                         inBusNumber,
    UInt32                         inNumberFrames,
    AudioBufferList                *ioData) {
    memset(ioData->mBuffers[0].mData, 0, ioData->mBuffers[0].mDataByteSize);
    n80sound *n80snd = (__bridge n80sound*)inRefCon;
    [n80snd callback:inNumberFrames ioData:ioData];

    return noErr;
}
//-----------------------------------------------------------------------------
//    Emulator Objective-C Class
//-----------------------------------------------------------------------------
@implementation n80sound

-(id)initWithN80:(n80em*)n80em {
    if (self = [super init]) {
        self->n80em = n80em;
        // init member variables
        self->gOutputUnit = NULL;
        // lock for n80 sync
        self->sync = [[NSRecursiveLock alloc] init];

        // beep/sound bank init
        memset(&beep_bank, 0, sizeof(beep_bank));
        memset(&sound_bank, 0, sizeof(sound_bank));
        //beep_bank.current.state = 1;
        //beep_bank.current.ticks = n80device.param.sync_ticks * 4;
        // CPU Ticks変換係数: frames = (44100/3993600*Ts)
        n80lx.sampling_rate = (int)SAMPLE_RATE;
        n80lx.frame_keisu = (double) n80lx.sampling_rate / (double) n80device.param.hz;

        if (![self prepareAudio])
            return nil;
    }
    return self;
}

-(void)dealloc {
    NSLog(@"n80sound dealloc");
    [self reqDispose];
}

-(N80LX const *)getN80 {
    return &n80lx;
}

-(void)lock {
    [self->sync lock];
}

-(void)unlock {
    [self->sync unlock];
}

-(void)suspend {
    if (gOutputUnit != NULL) {
        AudioOutputUnitStop(gOutputUnit);
        AudioUnitUninitialize(gOutputUnit);
        AudioComponentInstanceDispose(gOutputUnit);
        self->gOutputUnit = NULL;
    }
}

-(BOOL)resume {
    return [self prepareAudio];
}

- (void)reqDispose {
    // stop sound unit
    if (gOutputUnit != NULL) {
        AudioOutputUnitStop(gOutputUnit);
        AudioUnitUninitialize(gOutputUnit);
        AudioComponentInstanceDispose(gOutputUnit);
    }
    self->sync = NULL;
}
-(void)append_sound_ticks:(DWORD)T {
    // CPU STATE数をALSA frame数に変換して累積する
    double frames = n80lx.frame_keisu * T;
    [self lock];
    beep_bank.current.ticks     += frames;
    sound_bank[0].current.ticks += frames;
    sound_bank[1].current.ticks += frames;
    sound_bank[2].current.ticks += frames;
    [self unlock];
}
//-----------------------------------------------------------------------------
//    Initialize Emulator
//-----------------------------------------------------------------------------
- (BOOL)prepareAudio {
    OSStatus    err = noErr;
    int         sampleRate = SAMPLE_RATE;
    
    // Open the default output unit
    AudioComponentDescription desc;
    desc.componentType = kAudioUnitType_Output;
#if TARGET_OS_OSX
    desc.componentSubType = kAudioUnitSubType_DefaultOutput;
#else
	desc.componentSubType = kAudioUnitSubType_RemoteIO;
#endif
    desc.componentManufacturer = kAudioUnitManufacturer_Apple;
    desc.componentFlags = 0;
    desc.componentFlagsMask = 0;
    
    AudioComponent comp = AudioComponentFindNext(NULL, &desc);
    if (comp == NULL) {
        NSLog(@"AudioComponentFindNext error");
        return false;
    }
    err = AudioComponentInstanceNew(comp, &gOutputUnit);
    if (comp == NULL) {
        //NSLog(@"AudioComponentInstanceNew=%d", err);
        return false;
    }
    err = AudioUnitInitialize(gOutputUnit);
    
    // Set up a callback function to generate output to the output unit
    AURenderCallbackStruct callback;
    callback.inputProc = audioCallback;
    callback.inputProcRefCon = (__bridge void*)self;
    
    err = AudioUnitSetProperty (gOutputUnit,
                                kAudioUnitProperty_SetRenderCallback,
                                //kAudioOutputUnitProperty_EnableIO,
                                kAudioUnitScope_Input,        //initial value is "kAudioUnitScope_Input". is it ok?
                                0,
                                &callback,
                                sizeof(callback));
    
    AudioStreamBasicDescription streamFormat;
    streamFormat.mSampleRate = sampleRate;                    //    the sample rate of the audio stream
    streamFormat.mFormatID = kAudioFormatLinearPCM;            //    the specific encoding type of audio stream
#ifdef __BIG_ENDIAN__
    streamFormat.mFormatFlags =        kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsBigEndian
    | kLinearPCMFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;        //    16bit int
#else
    streamFormat.mFormatFlags =        kLinearPCMFormatFlagIsSignedInteger
    | kLinearPCMFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;        //    16bit int
#endif
    streamFormat.mBytesPerPacket = 2;
    streamFormat.mFramesPerPacket = 1;
    streamFormat.mBytesPerFrame = 2;
    streamFormat.mChannelsPerFrame = 1;
    streamFormat.mBitsPerChannel = 16;
    
    err = AudioUnitSetProperty (gOutputUnit,
                                kAudioUnitProperty_StreamFormat,
                                kAudioUnitScope_Input,
                                0,
                                &streamFormat,
                                sizeof(AudioStreamBasicDescription));
        
    Float64 outSampleRate;
    UInt32 size = sizeof(Float64);
    
    err = AudioUnitGetProperty (gOutputUnit,
                                kAudioUnitProperty_SampleRate,
                                kAudioUnitScope_Output,
                                0,
                                &outSampleRate,
                                &size);
    
    err = AudioOutputUnitStart (gOutputUnit);
    
    return true;
}


-(void)cpu_reset {
    //[syncSound lock];                    // reset sound tick buffer
    [self lock];
    beep_bank.bank = 0;
    beep_bank.banks[0].count = 0;
    beep_bank.banks[0].ticks_sum = 0;
    beep_bank.banks[1].count = 0;
    beep_bank.banks[1].ticks_sum = 0;
    beep_bank.current.state = 0;
    beep_bank.current.ticks = 0;
    for (int i = 0; i < 3; ++i) {
        sound_bank[i].bank = 0;
        sound_bank[i].banks[0].count = 0;
        sound_bank[i].banks[0].ticks_sum = 0;
        sound_bank[i].banks[1].count = 0;
        sound_bank[i].banks[1].ticks_sum = 0;
        sound_bank[i].current.state = 0;
        sound_bank[i].current.ticks = 0;
    }
    //[syncSound unlock];
    [self unlock];
}

-(void)beep_on {
    [self lock];
    if (beep_bank.current.state == 0) {
        append_tick_to_array(&beep_bank, 1);
    }
    [self unlock];
}

-(void)beep_off {
    [self lock];
    if (beep_bank.current.state == 1) {
        append_tick_to_array(&beep_bank, 0);
    }
    [self unlock];
}

-(void)sound_on:(USHORT)freq ch:(int)ch {
    [self lock];
    if (sound_bank[ch].current.state != freq)
        for (int i = 0; i < 3; ++i) {
            USHORT f = ch == i ? freq : sound_bank[i].current.state;
            append_tick_to_array(&sound_bank[i], f);
        }
    [self unlock];
}

-(void)sound_off {
    [self lock];
    append_tick_to_array(&sound_bank[0], 0);
    append_tick_to_array(&sound_bank[1], 0);
    append_tick_to_array(&sound_bank[2], 0);
    [self unlock];
}

static inline void sample_overlay(int16_t *pcm, int val) {
    int32_t v = *pcm;
    v += val;
    if (v > SHRT_MAX)
        v = SHRT_MAX;
    else if (v < SHRT_MIN)
        v = SHRT_MIN;
    *pcm++ = (int16_t)v;
}

static inline void addTk(LPSAMPLETsARR arr, double ticks, DWORD state) {
    arr->arr[arr->count].ticks = ticks;
    arr->arr[arr->count].state = state;
    ++arr->count;
    arr->ticks_sum += ticks;
}
static inline void appendTk(LPSAMPLETsARR arr, double ticks, DWORD state) {
    if (arr->count > 0 && arr->arr[arr->count-1].state == state)
        arr->arr[arr->count-1].ticks += ticks;
    else {
        arr->arr[arr->count].ticks = ticks;
        arr->arr[arr->count].state = state;
        ++arr->count;
    }
    arr->ticks_sum += ticks;
}

static void exchange_bank(DWORD au_frames, int *_beep_bank, int *_sound_bank) {
    int beep_primary_bank, sound_primary_bank;

    beep_primary_bank    = beep_bank.bank;
    sound_primary_bank    = sound_bank[0].bank;
    int beep_sub_bank    = beep_primary_bank == 0 ? 1 : 0;
    int sound_sub_bank    = sound_primary_bank == 0 ? 1 : 0;
    // clear sub bank
    beep_bank.banks[beep_sub_bank].count          = 0;
    beep_bank.banks[beep_sub_bank].ticks_sum      = 0;
    sound_bank[0].banks[sound_sub_bank].count     = 0;
    sound_bank[0].banks[sound_sub_bank].ticks_sum = 0;
    sound_bank[1].banks[sound_sub_bank].count     = 0;
    sound_bank[1].banks[sound_sub_bank].ticks_sum = 0;
    sound_bank[2].banks[sound_sub_bank].count     = 0;
    sound_bank[2].banks[sound_sub_bank].ticks_sum = 0;

    static double total_ticks = 0;
    static double total_rollover_ticks = 0;
    static double total_frames = 0;
    static double discard_frames = 0;
    static double loop_count = 0;
    //---- Beep ---------------------------------------------------------------
    LPSAMPLETsARR arr1 = &beep_bank.banks[beep_primary_bank];
    LPSAMPLETsARR arr2 = &beep_bank.banks[beep_sub_bank];

    // rack of frames
    if (arr1->ticks_sum < au_frames) {
        double diff = au_frames - arr1->ticks_sum;
        advance_tick_to_array(&beep_bank, diff);
    }
    //pick up need frames
    if (arr1->ticks_sum > au_frames) {
        LPSAMPLETsARR    _arr = &beep_copy_arr;
        _arr->ticks_sum    = 0;
        _arr->count        = 0;
        bool overflow = false;
        for (int i = 0; i < arr1->count; ++i) {
            double ticks = arr1->arr[i].ticks;
            DWORD  state = arr1->arr[i].state;
            if (!overflow) {
                double frames = _arr->ticks_sum + ticks;
                if (frames > au_frames) {
                    double over = frames - au_frames;
                    ticks -= over;

                    appendTk(arr2, over, state);
                    overflow = true;
                }
                appendTk(_arr, ticks, state);
            } else {
                appendTk(arr2, ticks, state);
            }
        }
        beep_primary_bank = -1;
    }
    //---- summary ----------------------------------------------------------------
    loop_count += 1;
    total_frames += au_frames;
    total_ticks += arr1->ticks_sum;
    total_rollover_ticks += arr2->ticks_sum;
    if (total_frames > SAMPLE_RATE*2) {
        //double percent = beep_frames / (total_frames) * 100;
        double percent1 = (total_ticks + total_rollover_ticks) / total_frames * 100;
        double percent2 = total_ticks / total_frames * 100;
        double percent3 = total_rollover_ticks / total_frames * 100;
        int ave_frames = (int)(total_frames / loop_count);
        double datas = (total_ticks + total_rollover_ticks) / loop_count;
        printf("sound ave=%d rcvdata = %3.2f(use:%3.2f, rollover=%3.2f) datas=%.0f\n", ave_frames, percent1, percent2, percent3, datas);
/*
        double ave_sync = total_frames / total_ticks * 100;
        double ave_frames = (total_frames + discard_frames) / loop_count;
        double ave_discard = discard_frames / loop_count;
        printf("sound percent = %3.2f(%3.2f)  frame = %8.1f discard = %4.2f\n", percent, ave_sync, ave_frames, ave_discard);
*/
        total_ticks = 0;
        total_frames = 0;
        //beep_frames = 0;
        total_rollover_ticks = 0;
        discard_frames = 0;
        loop_count = 0;
    }
    //---- PCG ----------------------------------------------------------------
    LPSAMPLETsARR arr10 = &sound_bank[0].banks[sound_primary_bank];        // current
    LPSAMPLETsARR arr20 = &sound_bank[0].banks[sound_sub_bank];            // next
    LPSAMPLETsARR arr11 = &sound_bank[1].banks[sound_primary_bank];        // current
    LPSAMPLETsARR arr21 = &sound_bank[1].banks[sound_sub_bank];            // next
    LPSAMPLETsARR arr12 = &sound_bank[2].banks[sound_primary_bank];        // current
    LPSAMPLETsARR arr22 = &sound_bank[2].banks[sound_sub_bank];            // next
    
    // rack of frames
    if (arr10->ticks_sum < au_frames) {
        double diff = au_frames - arr10->ticks_sum;
        advance_tick_to_array(&sound_bank[0], diff);
        advance_tick_to_array(&sound_bank[1], diff);
        advance_tick_to_array(&sound_bank[2], diff);
    }
    //pick up need frames
    if (arr10->ticks_sum > au_frames) {
        LPSAMPLETsARR _arr0 = &sound_copy_arr[0];
        LPSAMPLETsARR _arr1 = &sound_copy_arr[1];
        LPSAMPLETsARR _arr2 = &sound_copy_arr[2];
        _arr0->ticks_sum    = 0;
        _arr0->count        = 0;
        _arr1->ticks_sum    = 0;
        _arr1->count        = 0;
        _arr2->ticks_sum    = 0;
        _arr2->count        = 0;
        bool overflow = false;
        for (int i = 0; i < arr10->count; ++i) {
            double ticks = arr10->arr[i].ticks;
            DWORD  state0 = arr10->arr[i].state;
            DWORD  state1 = arr11->arr[i].state;
            DWORD  state2 = arr12->arr[i].state;
            if (!overflow) {
                double frames = _arr0->ticks_sum + ticks;
                if (frames > au_frames) {
                    double over = frames - au_frames;
                    ticks -= over;

                    addTk(arr20, over, state0);
                    addTk(arr21, over, state1);
                    addTk(arr22, over, state2);
                    overflow = true;
                }
                addTk(_arr0, ticks, state0);
                addTk(_arr1, ticks, state1);
                addTk(_arr2, ticks, state2);
            } else {
                addTk(arr20, ticks, state0);
                addTk(arr21, ticks, state1);
                addTk(arr22, ticks, state2);
            }
        }
        sound_primary_bank = -1;
    }
    // bank exchange
    beep_bank.bank = beep_sub_bank;
    for (int i = 0; i < 3; ++i) {
        sound_bank[i].bank = sound_sub_bank;
    }
    *_beep_bank  = beep_primary_bank;
    *_sound_bank = sound_primary_bank;
}

static int make_pcm(DWORD au_frames, int16_t *au_buffer, int _beep_bank, int _sound_bank) {
    int16_t        *pcm = au_buffer;
    int         write_frames = 0, rollover = 0;

    //---- Beep ---------------------------------------------------------------
    LPSAMPLETsARR arr = _beep_bank == -1 ? &beep_copy_arr : &beep_bank.banks[_beep_bank];
    double rate = au_frames < arr->ticks_sum ? au_frames / arr->ticks_sum : 1.0;
    float volume[au_frames];
    for (int i = 0; i < au_frames; ++i)    volume[i] = 0;

    double rs = 0.0, re = 0.0, remain = 0;    //real range in frames
    int is = 0, ie = 0, cnt = 0;            //integer range in frames
    for (int i = 0; i < arr->count; ++i) {
        re += arr->arr[i].ticks;
        is = floor(rs);
        ie = floor(re);
        cnt = ie - is + 1;                    //number of frames to effect
        remain = 0;
        if (cnt == 0)
            ;
        else if (cnt == 1) {
            if (is < au_frames) {
                volume[is] += (re - rs) * arr->arr[i].state;
                //++beep_frames;
            }
        } else if (cnt == 2) {
            if (is < au_frames) {
                volume[is] += (is + 1 - rs) * arr->arr[i].state;
            }
            if (ie < au_frames) {
                volume[ie] += (re - ie) * arr->arr[i].state;
                //++beep_frames;
            }
        } else {
            if (is < au_frames)
                volume[is] += (is + 1 - rs) * arr->arr[i].state;
            for (int j = is + 1; j < ie; ++j) {
                if (j < au_frames) {
                    volume[j] += arr->arr[i].state;
                    //++beep_frames;
                } else
                    remain += 1;
            }
            if (ie < au_frames) {
                volume[ie] += (re - ie) * arr->arr[i].state;
                //++beep_frames;
            }
        }
        rs = re;
        //discard_frames += remain;
        /*
        if (remain > 0) {
            arr2->arr[rollover].ticks = remain;
            arr2->arr[rollover].state = arr1->arr[i].state;
            arr2->ticks_sum += remain;
            arr2->count++;
            ++rollover;
        }*/
    }
    for (int i = 0; i < au_frames; ++i) {
        if ((n80lx.phase_adjuster % 40) < 20) {
            *pcm++ = floor(WAVETOP * volume[i]);
        } else {
            *pcm++ = floor(WAVEBOTTOM * volume[i]);
        }
        if (++n80lx.phase_adjuster > 1000000000)
            n80lx.phase_adjuster = 0;
    }
    //---- PCG ----------------------------------------------------------------
    pcm = au_buffer;
    (void)(write_frames = 0), rollover = 0;
    
    LPSAMPLETsARR    arr_0 = _sound_bank == -1 ? &sound_copy_arr[0] : &sound_bank[0].banks[_sound_bank];
    LPSAMPLETsARR    arr_1 = _sound_bank == -1 ? &sound_copy_arr[1] : &sound_bank[1].banks[_sound_bank];
    LPSAMPLETsARR    arr_2 = _sound_bank == -1 ? &sound_copy_arr[2] : &sound_bank[2].banks[_sound_bank];
    for (int i = 0; i < arr_0->count; ++i) {
        int frames = (int) ceil(arr_0->arr[i].ticks * rate);
        int remain = frames;
        int wave_frames0 = arr_0->arr[i].state > 0 ? floor((double) n80lx.sampling_rate / (double) arr_0->arr[i].state + 0.5) : 0;
        int wave_frames1 = arr_1->arr[i].state > 0 ? floor((double) n80lx.sampling_rate / (double) arr_1->arr[i].state + 0.5) : 0;
        int wave_frames2 = arr_2->arr[i].state > 0 ? floor((double) n80lx.sampling_rate / (double) arr_2->arr[i].state + 0.5) : 0;
        int half_wave_frames0 = wave_frames0 / 2;
        int half_wave_frames1 = wave_frames1 / 2;
        int half_wave_frames2 = wave_frames2 / 2;
        if (frames > 0) {
            for (int j = 0; j < frames; ++j) {
                // check if one period long over
                if (write_frames >= au_frames) {
                    break;
                }
                ++write_frames;
                --remain;
                int volume = 0;
                if (wave_frames0 > 0)    {    // ch.#0 sound on
                    if ((n80lx.phase_adjuster2 % wave_frames0) < half_wave_frames0) {
                        volume = WAVETOP;
                    } else {
                        volume = WAVEBOTTOM;
                    }
                }
                if (wave_frames1 > 0)    {    // ch.#1 sound on
                    if ((n80lx.phase_adjuster2 % wave_frames1) < half_wave_frames1) {
                        volume += WAVETOP;
                    } else {
                        volume += WAVEBOTTOM;
                    }
                }
                if (wave_frames2 > 0)    {    // ch.#2 sound on
                    if ((n80lx.phase_adjuster2 % wave_frames2) < half_wave_frames2) {
                        volume += WAVETOP;
                    } else {
                        volume += WAVEBOTTOM;
                    }
                }
                sample_overlay(pcm++, volume);
                if (++n80lx.phase_adjuster2 > 100000000)
                    n80lx.phase_adjuster2 = 0;
            }
        }
        /*
        if (remain > 0) {
            arr20->arr[rollover].ticks = remain;
            arr20->arr[rollover].state = arr10->arr[i].state;
            arr20->ticks_sum += remain;
            arr20->count++;
            arr21->arr[rollover].ticks = remain;
            arr21->arr[rollover].state = arr11->arr[i].state;
            arr21->ticks_sum += remain;
            arr21->count++;
            arr22->arr[rollover].ticks = remain;
            arr22->arr[rollover].state = arr12->arr[i].state;
            arr22->ticks_sum += remain;
            arr22->count++;
            ++rollover;
        }*/
    }
    return write_frames;
}

-(void)callback:(UInt32)inNumberFrames ioData:(AudioBufferList *)ioData {
    int            beep_bank, sound_bank;
    int16_t        *pcm = ioData->mBuffers[0].mData;
    DWORD        au_frames = ioData->mBuffers[0].mDataByteSize / 2;

    [self lock];
    exchange_bank(au_frames, &beep_bank, &sound_bank);
    [self unlock];
    make_pcm(au_frames, pcm, beep_bank, sound_bank);
}

@end

